"
COPYRIGHT (c) 1989 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

DeviceWorkstation subclass:#XWorkstation
       instanceVariableNames:'screen
                              hasShapeExtension hasFaxExtension hasShmExtension
                              hasDPSExtension hasSaveUnder ignoreBackingStore
                              blackpixel whitepixel redMask greenMask blueMask
                              ProtocolsAtom DeleteWindowAtom SaveYourselfAtom
                              QuitAppAtom
                              listOfXFonts eventRootX eventRootY buttonsPressed
                              displayName eventTrace
                              dispatchingExpose
                              rgbVisual'
       classVariableNames:   'KeyPressMask PointerMotionMask
                              ButtonPressMask ButtonReleaseMask ButtonMotionMask
                              ExposureMask FocusChangeMask StructureNotifyMask
                              EnterWindowMask LeaveWindowMask
                              BitButton1 BitButton2 BitButton3'
       poolDictionaries:''
       category:'Interface-Graphics'
!

XWorkstation comment:'

COPYRIGHT (c) 1989 by Claus Gittinger
              All Rights Reserved

this class provides the interface to X11

$Header: /cvs/stx/stx/libview/XWorkstation.st,v 1.15 1994-05-12 10:30:43 claus Exp $
written spring/summer 89 by claus
'!

%{
/* x does a typedef Time - I need Object Time ... */

#define Time XTime

#ifdef memset
# undef memset
#endif

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#define TWOBYTESTRINGS

/*
 * shape support works; enabled by -DSHAPE in makefile
 * see RoundClock and RoundGlobe examples
 */
#ifdef SHAPE
# include <X11/extensions/shape.h>
#endif

/*
 * shared memory extension currently not supported
 * (only query is implemented)
 */
#ifdef SHM
# include <X11/extensions/XShm.h>
#endif

/*
 * this is a non-standard fax-image decompression extension
 * (only found in megascan servers ported by myself)
 */
#ifdef FAX
# include <X11/extensions/XFAXImage.h>
#endif

/*
 * when I have more time to check it out, Ill support display-PS
 */
#ifdef DPS
# ifdef sgi
#  include <X11/extensions/XDPS.h>
#  include <X11/extensions/XDPSlib.h>
#  include <X11/extensions/dpsXclient.h>
# else
#  include <DPS/XDPS.h>
#  include <DPS/XDPSlib.h>
#  include <DPS/dpsXclient.h>
# endif
#endif

static XEvent ev;

/*
 * a private error handler
 */
static char lastErrorMsg[80] = "";
static unsigned lastRequestCode = 0;
static unsigned lastMinorCode = 0;
static unsigned lastResource = 0;

/*
 * catch X-errors and forward as errorInterrupt,
 * (which itself raises an exceptionSignal)
 */
__XErrorHandler__(dpy, event)
    Display *dpy;
    XErrorEvent *event;
{
    XGetErrorText(dpy, event->error_code, lastErrorMsg, 80);
    if (lastErrorMsg[0] == '\0') {
        sprintf(lastErrorMsg, "code: %d", event->error_code);
    }
    lastRequestCode = event->request_code;
    lastMinorCode = event->minor_code;
    lastResource = event->resourceid;

    fprintf(stderr, "x-error cought maj=%d (0x%x) min=%d (0x%x) resource=%x\n",
                    event->request_code, event->request_code, 
                    event->minor_code, event->minor_code, event->resourceid);
    fprintf(stderr, "        (%s)\n", lastErrorMsg);

    errorInterrupt();
    return 0;
}

/*
 * added since RootWindow-macro is not suficient
 * when virtual root-windows are involved (i.e. tvtwm)
 */
static Window
getRootWindow(dpy, screen)
    Display *dpy;
{
    Window root;
    Atom vRootAtom = None;
    int i;
    Window rootReturn, parentReturn;
    Window* children;
    unsigned int numChildren;

    root = RootWindow(dpy, screen);
    /*
     * on IRIS, this creates a badwindow error - why ?
     * children contains a funny window (000034)
     */
#ifndef IRIS
    if (XQueryTree(dpy, root, &rootReturn, &parentReturn, &children, &numChildren)) {
        vRootAtom = XInternAtom(dpy, "__SWM_VROOT", True );
        if (vRootAtom != None) {
            for (i=0; i < numChildren; i++) {
                Atom actual_type;
                int actual_format;
                unsigned long nitems, bytesafter;
                Window* newRoot = (Window*) 0;

                if (children[i]) {
                    if (XGetWindowProperty(dpy, children[i], vRootAtom,
                                           0L, 1L, False, XA_WINDOW,
                                           &actual_type, &actual_format, &nitems, &bytesafter,
                                           (unsigned char**) &newRoot) == Success && newRoot) {
                        root = *newRoot;
                        break;
                    }
                }
            }
            if (children) XFree( children );
        }
    }
#endif
    return root;
}

/*
 * some defines - tired of typing ...
 */
#if defined(hpux) && defined(POSITIVE_ADDRESSES)
# define _DisplayVal(o)       (Display *)((int)(o) & ~TAG_INT)
# define _WindowVal(o)        (Window)((int)(o) & ~TAG_INT)
# define _PixmapVal(o)        (Pixmap)((int)(o) & ~TAG_INT)
# define _GCVal(o)            (GC)((int)(o) & ~TAG_INT)
# define _CursorVal(o)        (Cursor)((int)(o) & ~TAG_INT)
# define _FontVal(o)          (XFontStruct *)((int)(o) & ~TAG_INT)
# define _AtomVal(o)          (Atom)((int)(o) & ~TAG_INT)
# define _DPSContextVal(o)    (DPSContext)((int)(o) & ~TAG_INT)
#else
# define _DisplayVal(o)       (Display *)(_intVal(o))
# define _WindowVal(o)        (Window)(_intVal(o))
# define _PixmapVal(o)        (Pixmap)(_intVal(o))
# define _GCVal(o)            (GC)(_intVal(o))
# define _CursorVal(o)        (Cursor)(_intVal(o))
# define _FontVal(o)          (XFontStruct *)(_intVal(o))
# define _AtomVal(o)          (Atom)(_intVal(o))
# define _DPSContextVal(o)    (DPSContext)(_intVal(o))
#endif

#define myDpy _DisplayVal(_INST(displayId))

#ifndef THISCONTEXT_IN_REGISTER
# define BLOCKINTERRUPTS()      /* */
# define UNBLOCKINTERRUPTS()    /* */
#endif

%}

!XWorkstation class methodsFor:'initialization'!

initialize
    self initializeDisplayConstants.
    self initializeConstants.
!

initializeDisplayConstants
    "initialize some display constants"

%{  /* NOCONTEXT */

    XWorkstation_KeyPressMask = _MKSMALLINT(KeyPressMask);
    XWorkstation_ButtonPressMask = _MKSMALLINT(ButtonPressMask);
    XWorkstation_ButtonReleaseMask = _MKSMALLINT(ButtonReleaseMask);
    XWorkstation_ButtonMotionMask = _MKSMALLINT(Button1MotionMask);
    XWorkstation_PointerMotionMask = _MKSMALLINT(PointerMotionMask);
    XWorkstation_ExposureMask = _MKSMALLINT(ExposureMask);
    XWorkstation_FocusChangeMask = _MKSMALLINT(FocusChangeMask);
    XWorkstation_EnterWindowMask = _MKSMALLINT(EnterWindowMask);
    XWorkstation_LeaveWindowMask = _MKSMALLINT(LeaveWindowMask);
    XWorkstation_StructureNotifyMask = _MKSMALLINT(StructureNotifyMask);
    XWorkstation_BitButton1 = _MKSMALLINT(Button1Mask);
    XWorkstation_BitButton2 = _MKSMALLINT(Button2Mask);
    XWorkstation_BitButton3 = _MKSMALLINT(Button3Mask);
%}
! !

!XWorkstation class methodsFor:'error handling'!

requestCodeOfLastError
%{  /* NOCONTEXT */

    RETURN ( _MKSMALLINT(lastRequestCode) );
%}
!

minorCodeOfLastError
%{  /* NOCONTEXT */

    RETURN ( _MKSMALLINT(lastMinorCode) );
%}
!

resourceIdOfLastError
%{  /* NOCONTEXT */

    RETURN ( _MKSMALLINT(lastResource) );
%}
!

lastError
    "return the last X-error string - 
     when buffering is on, this may be
     an error for a long-ago operation"
%{

    RETURN ( _MKSTRING(lastErrorMsg, __context) );
%}
! !

!XWorkstation methodsFor:'initialize / release'!

initializeFor:aDisplayName
    "initialize the receiver for a connection to an X-Server;
     the argument, aDisplayName may be nil (for the default server from
     DISPLAY-variable) or the name of the server as hostname:number"

%{
    int scr;
    Display *dpy;
    Visual *visual;
    XVisualInfo viproto;
    XVisualInfo *vip;                   /* retured info */
    int maxRGBDepth;
    int rgbRedMask, rgbGreenMask, rgbBlueMask;
    int rgbVisualID;
    int nvi, i;
    int shapeEventBase, shapeErrorBase;
    int shmEventBase, shmErrorBase;
    int faxEventBase, faxErrorBase;
    char *type;
    int dummy;

    if (_INST(displayId) != nil) {
        RETURN ( self );
    }

    BLOCKINTERRUPTS();
    if (aDisplayName != nil)
        dpy = XOpenDisplay((char *)_stringVal(aDisplayName));
    else
        dpy = XOpenDisplay(NULL);

    if (dpy) {
#ifdef SUPERDEBUG
        XSynchronize(dpy, 1);
#endif
        _INST(screen) = _MKSMALLINT(scr = DefaultScreen(dpy));
        _INST(displayId) = MKOBJ((int)dpy);
        _INST(depth) = _MKSMALLINT(DisplayPlanes(dpy, scr));
        _INST(ncells) = _MKSMALLINT(DisplayCells(dpy, scr));
        _INST(width) = _MKSMALLINT(DisplayWidth(dpy, scr));
        _INST(height) = _MKSMALLINT(DisplayHeight(dpy, scr));
        _INST(widthMM) = _MKSMALLINT(DisplayWidthMM(dpy, scr));
        _INST(heightMM) = _MKSMALLINT(DisplayHeightMM(dpy, scr));
        _INST(blackpixel) = _MKSMALLINT(BlackPixel(dpy, scr));
        _INST(whitepixel) = _MKSMALLINT(WhitePixel(dpy, scr));
#ifdef SHAPE
        if (XShapeQueryExtension(dpy, &shapeEventBase, &shapeErrorBase))
            _INST(hasShapeExtension) = true;
        else
#else
        _INST(hasShapeExtension) = false;
#endif
#ifdef SHM
        if (XShmQueryExtension(dpy, &shmEventBase, &shmErrorBase))
            _INST(hasShmExtension) = true;
        else
#else
        _INST(hasShmExtension) = false;
#endif
#ifdef FAX
        if (XFAXImageQueryExtension(dpy, &faxEventBase, &faxErrorBase))
            _INST(hasFaxExtension) = true;
        else
#else
        _INST(hasFaxExtension) = false;
#endif
#ifdef DPS
        if (XQueryExtension(dpy, "DPSExtension", &dummy, &dummy, &dummy))
            _INST(hasDPSExtension) = true;
        else
#else
        _INST(hasDPSExtension) = false;
#endif
        /*
         * this is a kludge around a bug in the X11/NeWS server,
         * which does not correctly handle saveUnder
         */
        if (strncmp(XServerVendor(dpy), "X11/NeWS", 8) == 0) {
            _INST(hasSaveUnder) = false;
        } else
            _INST(hasSaveUnder) = true;

        /*
         * look for RGB visual
         */
        nvi = 0;
        viproto.screen = scr;
        vip = XGetVisualInfo (dpy, VisualScreenMask, &viproto, &nvi);
        maxRGBDepth = 0;
        for (i = 0; i < nvi; i++) {
            switch (vip[i].class) {
                case TrueColor:
                    if (vip[i].depth > maxRGBDepth) {
                        maxRGBDepth = vip[i].depth;
                        rgbRedMask = vip[i].red_mask;
                        rgbGreenMask = vip[i].green_mask;
                        rgbBlueMask = vip[i].blue_mask;
                        rgbVisualID = vip[i].visualid;
                    }
                    break;
            }
        }
        if (vip) XFree ((char *) vip);

        if (maxRGBDepth) {
            _INST(rgbVisual) = _MKSMALLINT(rgbVisualID);
        }

        visual = DefaultVisualOfScreen(DefaultScreenOfDisplay(dpy));
        switch (visual->class) {
            case StaticGray:
                _INST(visualType) = @symbol(StaticGray);
                _INST(hasColors) = false;
                _INST(hasGreyscales) = true;
                _INST(monitorType) = @symbol(monochrome);
                break;
            case GrayScale:
                _INST(visualType) = @symbol(GrayScale);
                _INST(hasColors) = false;
                _INST(hasGreyscales) = true;
                _INST(monitorType) = @symbol(monochrome);
                break;
            case StaticColor:
                _INST(visualType) = @symbol(StaticColor);
                _INST(hasColors) = true;
                _INST(hasGreyscales) = true;
                _INST(monitorType) = @symbol(unknown);
                break;
            case PseudoColor:
                _INST(visualType) = @symbol(PseudoColor);
                _INST(hasColors) = true;
                _INST(hasGreyscales) = true;
                _INST(monitorType) = @symbol(unknown);
                break;
            case TrueColor:
                _INST(visualType) = @symbol(TrueColor);
                _INST(hasColors) = true;
                _INST(hasGreyscales) = true;
                _INST(monitorType) = @symbol(unknown);
                break;
            case DirectColor:
                _INST(visualType) = @symbol(DirectColor);
                _INST(hasColors) = true;
                _INST(hasGreyscales) = true;
                _INST(monitorType) = @symbol(unknown);
                break;
        }
        if (DisplayCells(dpy, scr) == 2) {
            _INST(hasColors) = false;
            _INST(hasGreyscales) = false;
            _INST(monitorType) = @symbol(monochrome);
        }
        _INST(bitsPerRGB) = _MKSMALLINT(visual->bits_per_rgb);
        _INST(redMask)   = _MKSMALLINT(visual->red_mask);
        _INST(greenMask) = _MKSMALLINT(visual->green_mask);
        _INST(blueMask)  = _MKSMALLINT(visual->blue_mask);

        XSetErrorHandler(__XErrorHandler__);
    }
    UNBLOCKINTERRUPTS();
%}
.
    displayId isNil ifTrue:[
        'cannot connect to Display.' errorPrintNewline.
        ^ nil
    ].
    dispatching := false.
    dispatchingExpose := false.
    isSlow := false.
    shiftDown := false.
    controlDown := false.
    metaDown := false.
    altDown := false.
    motionEventCompression := true.
    buttonsPressed := 0.
    displayName := aDisplayName.

    ProtocolsAtom := nil.
    DeleteWindowAtom := nil.
    SaveYourselfAtom := nil.
    QuitAppAtom := nil.
    self initializeKeyboardMap.

    ObjectMemory errorInterruptHandler:self class.
!

close
    "close down the connection to the X-server"

%{  /* NOCONTEXT */

    Display *dpy;

    if (_INST(displayId) != nil) {
        BLOCKINTERRUPTS();
        XCloseDisplay(myDpy);
        UNBLOCKINTERRUPTS();
        _INST(displayId) = nil;
    }
%}
!

reinitialize
    super reinitialize.
    dispatchingExpose := nil
! !

!XWorkstation methodsFor:'misc'!

setInputFocusTo:aWindowId

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId)) {
        BLOCKINTERRUPTS();
        XSetInputFocus(myDpy, _WindowVal(aWindowId), RevertToParent, CurrentTime);
        UNBLOCKINTERRUPTS();
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

unBuffered
    "make all drawing be sent immediately to the display"

%{  /* NOCONTEXT */

    BLOCKINTERRUPTS();
    XSynchronize(myDpy, 1);
    UNBLOCKINTERRUPTS();
%}
    "Display unbuffered"
!

buffered
    "buffer drawing - do not send it immediately to the display"

%{  /* NOCONTEXT */

    BLOCKINTERRUPTS();
    XSynchronize(myDpy, 0);
    UNBLOCKINTERRUPTS();
%}
    "Display buffered"
!

synchronizeOutput
    "send all buffered drawing to the display"

%{  /* NOCONTEXT */

    BLOCKINTERRUPTS();
    XSync(myDpy, 0);
    UNBLOCKINTERRUPTS();
%}
!

flushDpsContext:aDPSContext

%{  /* NOCONTEXT */
#ifdef DPS
    if (_isSmallInteger(aDPSContext)) {
        BLOCKINTERRUPTS();
        DPSFlushContext(MKDPSCONTEXT(aDPSContext));
        UNBLOCKINTERRUPTS();
        RETURN ( self );
    }
#endif
%}
.
    self primitiveFailed
!

beep:volumeInPercent
    "output an audible beep"
%{
    int volume;

    if (_isSmallInteger(volumeInPercent)) {
        /* stupid: X wants -100 .. 100 and calls this percent */
        volume = _intVal(volumeInPercent) * 2 - 100;
        if (volume < -100) volume = -100;
        else if (volume > 100) volume = 100;
        BLOCKINTERRUPTS();
        XBell(myDpy, volume);
        UNBLOCKINTERRUPTS();
    }
%}
!

beep
    "output an audible beep or bell"

    self beep:50
!

ignoreBackingStore:aBoolean
    "if the argument is true, the views backingStore setting will be ignored, and
     no backing store used - this can be used on servers where backing store is
     very slow or is broken (can be put into display-rc-file)"

    ignoreBackingStore := aBoolean
! !

!XWorkstation methodsFor:'accessing & queries'!

displayFileDescriptor
    "return the displays fileNumber - for select"

%{  /* NOCONTEXT */

    RETURN ( MKOBJ(ConnectionNumber(myDpy)) );
%}
!

serverVendor
    "return the X-server vendor string - this should normally not be of
     any interrest, but can be for special cases
     (to avoid bugs in certain implementations)"
%{

    RETURN ( _MKSTRING(XServerVendor(myDpy), __context) );
%}

    "
     Display serverVendor
    "
!

vendorRelease
    "return the X-servers vendor release - should normally not be of
     any interrest, but can be for special cases.
     (to avoid bugs in certain implementations)"
%{

    RETURN ( _MKSMALLINT(XVendorRelease(myDpy)) );
%}

    "
     Display vendorRelease
    "
!

protocolVersion
    "return the X-servers protocol version - should normally not be of
     any interrest"
%{
    RETURN ( _MKSMALLINT(XProtocolVersion(myDpy)) );
%}

    "
     Display protocolVersion
    "
!

hasShape
    "return true, if this workstation supports non-rectangular windows"

    ^ hasShapeExtension

    "
     Display hasShape
    "
!

hasShm
    "return true, if this workstation supports shared pixmaps"

    ^ hasShmExtension
!

hasFax
    "return true, if this workstation supports decompression of fax images"

    ^ hasFaxExtension
!

hasDPS
    "return true, if this workstation supports display postscript"

    ^ hasDPSExtension
!

supportedImageFormats
    "return an array with supported image formats; each array entry
     is another array, consisting of depth and bitsPerPixel values."

    |nFormats formatArray|
%{
    Display *dpy = myDpy;
    nFormats = _MKSMALLINT(dpy->nformats);
%}.
    formatArray := Array new:nFormats.
    1 to:nFormats do:[:i |
	|bitsPerPixelInfo depthInfo|
%{
	ScreenFormat *format;
        Display *dpy = myDpy;

        format = dpy->pixmap_format;
        format += (_intVal(i)-1);
        bitsPerPixelInfo = _MKSMALLINT(format->bits_per_pixel);
        depthInfo = _MKSMALLINT(format->depth);
%}.
	formatArray at:i put:(Array with:depthInfo with:bitsPerPixelInfo).
    ].
    ^ formatArray

    "
     Display supportedFormats
    "
!

blackpixel
    "return the colornumber of black"

    ^ blackpixel
!

whitepixel
    "return the colornumber of white"

    ^ whitepixel
!

bitButton1
    ^ BitButton1
!

bitButton2
    ^ BitButton2
!

viewIdFromPoint:aPoint in:windowId
    "given a point in rootWindow, return the viewId of the subview of windowId
     hit by this coordinate.
     return nil if no view was hit.
     - use to find window to drop objects after a cross-view drag"

%{  /* NOCONTEXT */

    Display *dpy = myDpy;
    int screen = _intVal(_INST(screen));
    OBJ xp, yp;
    int xpos, ypos;
    Window child_return;

    if (_isSmallInteger(windowId)
     && __isPoint(aPoint)) {
        xp = _point_X(aPoint);
        yp = _point_Y(aPoint);
        if (_isSmallInteger(xp) && _isSmallInteger(yp)) {
            BLOCKINTERRUPTS();
            XTranslateCoordinates(dpy,
                                  RootWindow(dpy, screen),
                                  _WindowVal(windowId),
                                  _intVal(xp), _intVal(yp), 
                                  &xpos, &ypos, &child_return);
            UNBLOCKINTERRUPTS();
            if (child_return) {
                RETURN ( MKOBJ(child_return) );
            }
        }
    }
%}
.
    ^ nil
!

translatePoint:aPoint from:windowId1 to:windowId2
    "given a point in window1, return the coordinate in window2
     - use to xlate points from a window to rootwindow"

    |x1 y1 x2 y2|

    x1 := aPoint x.
    y1 := aPoint y.
%{
    int xpos, ypos;
    Window w1, w2, child_return;
    int screen = _intVal(_INST(screen));
    Display *dpy = myDpy;

    if (_isSmallInteger(windowId1) && _isSmallInteger(windowId2)
     && _isSmallInteger(x1) && _isSmallInteger(y1)) {
        w1 = _WindowVal(windowId1);
        w2 = _WindowVal(windowId2);
#ifndef IRIS
        if ((w1 == RootWindow(dpy, screen))
         || (w2 == RootWindow(dpy, screen))) {
            if (w1 == RootWindow(dpy, screen)) {
                w1 = getRootWindow(dpy, screen);
            }
            if (w2 == RootWindow(dpy, screen)) {
                w2 = getRootWindow(dpy, screen);
            }
        }
#endif
        XTranslateCoordinates(dpy, w1, w2,
                              _intVal(x1), _intVal(y1), 
                              &xpos, &ypos, &child_return);
        x2 = _MKSMALLINT(xpos);
        y2 = _MKSMALLINT(ypos);
    }
%}
.
    ^ (x2 @ y2)
! !

!XWorkstation methodsFor:'bitmap/window creation'!

createFaxImageFromArray:data width:w height:h type:type k:k msbFirst:msbFirst
    "create a new faxImage in the workstation
     type: 0 -> uncompressed
           1 -> group3 1D (k is void)
           2 -> group3 2D
           3 -> group4 2D (k is void)"

    |nBytes|

    hasFaxExtension ifFalse:[^ nil].
    nBytes := data size.
%{
#ifdef FAX
    Display *dpy = myDpy;
    int screen = _intVal(_INST(screen));
    FAXImage newImage;
    int _type, _msb, _k = 0;
    unsigned char *bits;
    extern OBJ ByteArray;

    if (_isSmallInteger(w) && _isSmallInteger(h)
     && _isSmallInteger(type) && _isSmallInteger(k)
     && _isSmallInteger(nBytes)
     && __isByteArray(data)) {
        if (msbFirst == true)
            _msb = 1;
        else
            _msb = 0;
        _type = _intVal(type);
        if ((_type < 0) || (_type > 3)) {
            RETURN ( nil );
        }
        if (_type == 2) {
            if (! _isSmallInteger(k)) {
                RETURN ( nil );
            }
            _k = _intVal(k);
        }
        bits =  _ByteArrayInstPtr(data)->ba_element;
        newImage = XFAXImageCreateImage(dpy, _intVal(w), _intVal(h),
                                        _type, _k, _msb,
                                        _intVal(nBytes), bits);
        RETURN ( (newImage != (FAXImage)0) ? MKOBJ(newImage) : nil );
    }
#else
    RETURN ( nil );
#endif
%}
.
    self primitiveFailed.
    ^ nil
!

createBitmapWidth:w height:h
    "allocate a bitmap on the Xserver, the contents is undefined
     (i.e. random). Return a bitmap id or nil"

%{  /* NOCONTEXT */

    Display *dpy = myDpy;
    int screen = _intVal(_INST(screen));
    Pixmap newBitmap;

    if (_isSmallInteger(w) && _isSmallInteger(h)) {
        newBitmap = XCreatePixmap(dpy, RootWindow(dpy, screen),
                                       _intVal(w), _intVal(h), 1);
        RETURN ( (newBitmap != (Pixmap)0) ? MKOBJ(newBitmap) : nil );
    }
%}
.
    self primitiveFailed.
    ^ nil
!

createPixmapWidth:w height:h depth:d
    "allocate a pixmap on the Xserver, the contents is undefined
     (i.e. random). Return a bitmap id or nil"

%{  /* NOCONTEXT */

    Display *dpy = myDpy;
    int screen = _intVal(_INST(screen));
    Pixmap newBitmap;

    if (_isSmallInteger(w) && _isSmallInteger(h)) {
        newBitmap = XCreatePixmap(dpy, RootWindow(dpy, screen),
                                       _intVal(w), _intVal(h), _intVal(d));
        RETURN ( (newBitmap != (Pixmap)0) ? MKOBJ(newBitmap) : nil );
    }
%}
.
    self primitiveFailed.
    ^ nil
!

createBitmapFromFile:aString for:aForm
    |id w h|

%{
    Display *dpy = myDpy;
    int screen = _intVal(_INST(screen));
    Pixmap newBitmap;
    char *filename;
    unsigned b_width, b_height;
    int b_x_hot, b_y_hot;
    int status;

    if (__isString(aString) || __isSymbol(aString)) {
        filename = (char *)_stringVal(aString);

        BLOCKINTERRUPTS();
        status = XReadBitmapFile(dpy, RootWindow(dpy, screen),
                                 filename, &b_width, &b_height, &newBitmap,
                                 &b_x_hot, &b_y_hot);
        UNBLOCKINTERRUPTS();

        if (status == BitmapSuccess) {
            w = _MKSMALLINT(b_width);
            h = _MKSMALLINT(b_height);
            id = MKOBJ(newBitmap);
        }
    }
%}
.
    id notNil ifTrue:[
        aForm width:w.
        aForm height:h
    ].
    ^ id
!

primCreateBitmapFromArray:anArray width:w height:h

%{  /* UNLIMITEDSTACK */

    Display *dpy = myDpy;
    int screen = _intVal(_INST(screen));
    Pixmap newBitmap;
    unsigned int b_width, b_height;
    REGISTER unsigned char *cp;
    REGISTER unsigned char *pBits;
    unsigned char *b_bits, *allocatedBits;
    int index, row;
    REGISTER int col;
    unsigned bits;
    static char reverseBitTable[256];
    static firstCall = 1;
    int nBytes;
    unsigned char fastBits[10000];
    OBJ num, *op;
    int bytesPerRow;
    extern OBJ Array, ByteArray;

    if (firstCall) {
        for (index=0; index < 256; index++) {
            reverseBitTable[index] = 0;
            if (index & 128) reverseBitTable[index] |=   1;
            if (index &  64) reverseBitTable[index] |=   2;
            if (index &  32) reverseBitTable[index] |=   4;
            if (index &  16) reverseBitTable[index] |=   8;
            if (index &   8) reverseBitTable[index] |=  16;
            if (index &   4) reverseBitTable[index] |=  32;
            if (index &   2) reverseBitTable[index] |=  64;
            if (index &   1) reverseBitTable[index] |= 128;
        }
        firstCall = 0;
    }

    if (_isSmallInteger(w) && _isSmallInteger(h) && _isNonNilObject(anArray)) {
        b_width = _intVal(w);
        b_height = _intVal(h);
        bytesPerRow = (b_width + 7) / 8;
        nBytes = b_height * bytesPerRow;
        if (nBytes < sizeof(fastBits)) {
            cp = b_bits = fastBits;
            allocatedBits = 0;
        } else {
            cp = b_bits = allocatedBits = (unsigned char *) malloc(nBytes);
        }
        
        if (_qClass(anArray) == Array) {
            index = 1;
            op = &(_ArrayInstPtr(anArray)->a_element[index - 1]);
            for (row = b_height; row; row--) {
                for (col = bytesPerRow; col; col--) {
                    num = *op++;
                    if (! _isSmallInteger(num)) goto fail;
                    bits = _intVal(num);
                    *cp++ = reverseBitTable[bits];
                }
            }
        } else {
            if (_qClass(anArray) == ByteArray) {
                pBits = _ByteArrayInstPtr(anArray)->ba_element;
                for (col = b_height*bytesPerRow; col; col--) {
                    *cp++ = reverseBitTable[*pBits++];
                }
            } else {
                goto fail;
            }
        }

        BLOCKINTERRUPTS();
        newBitmap = XCreateBitmapFromData(dpy, RootWindow(dpy, screen),
                                               (char *)b_bits, 
                                               b_width, b_height);
        UNBLOCKINTERRUPTS();
fail: ;
        if (allocatedBits)
            free(allocatedBits);
        RETURN ( newBitmap ? MKOBJ(newBitmap) : nil );
    }
%}
!

createBitmapFromArray:anArray width:w height:h
    |bitmapId|

    bitmapId := self primCreateBitmapFromArray:anArray width:w height:h.
    bitmapId isNil ifTrue:[
        self primitiveFailed
    ].
    ^ bitmapId
!

destroyPixmap:aDrawableId

%{  /* NOCONTEXT */

    if (_isSmallInteger(aDrawableId)) {
        XFreePixmap(myDpy, _PixmapVal(aDrawableId));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

destroyFaxImage:aFaxImageId

%{  /* NOCONTEXT */

#ifdef FAX
    if (_isSmallInteger(aFaxImageId)) {
        XFAXImageFreeImage(myDpy, (FAXImage)_WindowVal(aFaxImageId));
        RETURN ( self );
    }
#endif
%}
.
    self primitiveFailed
!

rootWindowFor:aView
    |id|
%{
    int screen = _intVal(_INST(screen));
    Window root;

#ifdef IRIS
    root = RootWindow(myDpy, screen);
#else
    /*
     * on IRIS, this creates a badwindow error - why ?
     * children contains a funny window (000034)
     */
    /*
     * care for virtual root windows (tvtwm & friends)
     */
    Atom vRootAtom = None;
    int i;
    Window rootReturn, parentReturn;
    Window* children;
    unsigned int numChildren;

    root = RootWindow(myDpy, screen);
    if (XQueryTree(myDpy, root, &rootReturn, &parentReturn, &children, &numChildren)) {
        vRootAtom = XInternAtom(myDpy, "__SWM_VROOT", True );
        if (vRootAtom != None) {
            for (i=0; i < numChildren; i++) {
                Atom actual_type;
                int actual_format;
                unsigned long nitems, bytesafter;
                Window* newRoot = (Window*) 0;

                if (children[i]) {
                    if (XGetWindowProperty(myDpy, children[i], vRootAtom, 
                                           0L, 1L, False, XA_WINDOW,
                                           &actual_type, &actual_format, &nitems, &bytesafter,
                                           (unsigned char**) &newRoot) == Success && newRoot) {
                        root = *newRoot;
                        break;
                    }
                }
            }
            if (children) XFree( children );
        }
    }
#endif
    if (! root) {
        RETURN ( nil );
    }
    id = MKOBJ(root);
%}
.
    self addKnownView:aView withId:id.
    ^ id
!

createWindowFor:aView left:xpos top:ypos width:wwidth height:wheight

    |ext minWidth minHeight maxWidth maxHeight 
     bWidth bColor viewBg viewBgId wsuperView wsuperViewId wcreateOnTop 
     winputOnly wlabel wcursor wcursorId wicon wiconId windowId
     weventMask wiconView wiconViewId bitGravity viewGravity vBgColor
     vBgForm deepForm preferedVisual preferedDepth|

    ext := aView minExtent.
    ext notNil ifTrue:[
        minWidth := ext x.
        minHeight := ext y
    ].
    ext := aView maxExtent.
    ext notNil ifTrue:[
        maxWidth := ext x.
        maxHeight := ext y
    ].

    bWidth := aView borderWidth.
    bColor := aView borderColor.
    bColor notNil ifTrue:[
        bColor := bColor colorId
    ].

    viewBg := aView viewBackground.
    viewBg notNil ifTrue:[
        viewBg isColor ifTrue:[
            viewBg := viewBg colorId.
        ] ifFalse:[
            viewBg := nil.
"
            vBgForm := viewBg ditherForm.
            vBgForm notNil ifTrue:[
                (vBgForm depth ~~ depth) ifTrue:[
                    deepForm := Form width:(vBgForm width)
                                    height:(vBgForm height)
                                     depth:depth
                                        on:self.
                    deepForm paint:viewBg.
                    deepForm fillRectangleX:0 y:0 width:(vBgForm width)
                                                 height:(vBgForm height).
                    viewBgId := deepForm id
                ] ifFalse:[
                    viewBgId := vBgForm id
                ]
            ]
"
        ]
    ].

    wsuperView := aView superView.
    wsuperView notNil ifTrue:[
        wsuperViewId := wsuperView id
    ].
    wcreateOnTop := aView createOnTop.
    winputOnly := aView inputOnly.
    wlabel := aView label.
    wcursor := aView cursor.
    wcursor isNil ifTrue:[
        'cursor nil - defaulted' errorPrintNewline
    ] ifFalse:[
        wcursorId := wcursor id
    ].

    wicon := aView icon.
    wicon notNil ifTrue:[
        wiconId := wicon id
    ].
    wiconView := aView iconView.
    wiconView notNil ifTrue:[
        wiconViewId := wiconView id
    ].
    weventMask := aView eventMask.

    bitGravity := aView bitGravity.
    viewGravity := aView viewGravity.
    preferedVisual := aView preferedVisual.
    preferedDepth := aView preferedDepth.

%{
    Display *dpy = myDpy;
    int screen = _intVal(_INST(screen));
    Visual visual;
    XGCValues xgcv;
    XSetWindowAttributes xswa;
    XSizeHints sizehints;
    XWMHints wmhints;
    int bw = 0;
    int bd, bg;
    Window newWindow, parentWindow;
    char *windowName;
    XFontStruct *f;
    Pixmap backPixmap = (Pixmap)0, iconBitmap = (Pixmap)0;
    int flags = 0, depth, ioClass;
    extern OBJ Rectangle, Form;
    Window iconWindow;
    Atom WmDeleteWindowAtom, WmSaveYourselfAtom, WmProtocolsAtom;
    Atom WmQuitAppAtom;
    Atom atoms[3];
    int atomCount = 0, isTopWindow = 0;

    sizehints.flags = 0;
    sizehints.width = 100;
    sizehints.height = 100;
    sizehints.x = 0;
    sizehints.y = 0;

    if (_isSmallInteger(wwidth) && _isSmallInteger(wheight)) {
        sizehints.flags |= PSize;
        sizehints.width = _intVal(wwidth);
        sizehints.height = _intVal(wheight);
    }
    if (_isSmallInteger(xpos) && _isSmallInteger(ypos)) {
        sizehints.flags |= PPosition;
        sizehints.x = _intVal(xpos);
        sizehints.y = _intVal(ypos);
    }
    if (_isSmallInteger(minWidth) && _isSmallInteger(minHeight)) {
        sizehints.flags |= PMinSize;
        sizehints.min_width = _intVal(minWidth);
        sizehints.min_height = _intVal(minHeight);
    }
    if (_isSmallInteger(maxWidth) && _isSmallInteger(maxHeight)) {
        sizehints.flags |= PMaxSize;
        sizehints.max_width = _intVal(maxWidth);
        sizehints.max_height = _intVal(maxHeight);
    }

    bg = WhitePixel(dpy, screen);
    if (_isSmallInteger(viewBg)) {
        bg = _intVal(viewBg);
    } else if (_isSmallInteger(viewBgId)) {
        backPixmap = _PixmapVal(viewBgId);
    }

    if (_isSmallInteger(bWidth)) {
        bw = _intVal(bWidth);
    }

    if (_isSmallInteger(bColor)) {
        bd = _intVal(bColor);
    } else {
        bd = BlackPixel(dpy, screen);
    }

    if (_isSmallInteger(wsuperViewId)) {
        parentWindow = _WindowVal(wsuperViewId);
    } else {
        parentWindow = RootWindow(dpy, screen);
        isTopWindow = 1;
    }

    if (bitGravity == @symbol(NorthWest)) {
        xswa.bit_gravity = NorthWestGravity;
        flags |= CWBitGravity;
    }

    if (viewGravity != nil) {
        if (viewGravity == @symbol(NorthWest)) {
            xswa.win_gravity = NorthWestGravity;
            flags |= CWWinGravity;
        }
        if (viewGravity == @symbol(NorthEast)) {
            xswa.win_gravity = NorthEastGravity;
            flags |= CWWinGravity;
        }
        if (viewGravity == @symbol(SouthWest)) {
            xswa.win_gravity = SouthWestGravity;
            flags |= CWWinGravity;
        }
        if (viewGravity == @symbol(SouthEast)) {
            xswa.win_gravity = SouthEastGravity;
            flags |= CWWinGravity;
        }
        if (viewGravity == @symbol(Center)) {
            xswa.win_gravity = CenterGravity;
            flags |= CWWinGravity;
        }
        if (viewGravity == @symbol(North)) {
            xswa.win_gravity = NorthGravity;
            flags |= CWWinGravity;
        }
        if (viewGravity == @symbol(South)) {
            xswa.win_gravity = SouthGravity;
            flags |= CWWinGravity;
        }
        if (viewGravity == @symbol(West)) {
            xswa.win_gravity = WestGravity;
            flags |= CWWinGravity;
        }
        if (viewGravity == @symbol(East)) {
            xswa.win_gravity = EastGravity;
            flags |= CWWinGravity;
        }
    }

    if (wcreateOnTop == true)
        xswa.override_redirect = 1;
    else 
        xswa.override_redirect = 0;

    if (winputOnly == true) 
        ioClass = InputOnly;
    else 
        ioClass = InputOutput;

    if (_isSmallInteger(weventMask)) {
        xswa.event_mask = _intVal(weventMask);
    } else {
        xswa.event_mask = 0;
    }

    /* these may not be disabled */
    xswa.event_mask |= ExposureMask | StructureNotifyMask;
    if (ioClass == InputOnly) {
        bw = 0;
        depth = 0;
        flags |= CWEventMask;
    } else {
        depth = DefaultDepth(dpy,screen);
        flags |= CWEventMask | CWBorderPixel | CWOverrideRedirect;

        if (backPixmap != (Pixmap)0) {
            xswa.background_pixmap = backPixmap;
            flags |= CWBackPixmap;
        } else {
            xswa.background_pixel = bg;
            flags |= CWBackPixel;
        }
        xswa.border_pixel = bd;
    }

    visual.visualid = CopyFromParent;
    if (_isSmallInteger(preferedDepth)) {
        depth = _intVal(preferedDepth);
    }
    if (preferedVisual != nil) {
        XVisualInfo vi;
        int cls;

        if (preferedVisual == @symbol(StaticGray))
            cls = StaticGray;
        else if (preferedVisual == @symbol(GrayScale)) 
            cls = StaticGray;
        else if (preferedVisual == @symbol(StaticColor)) 
            cls = StaticColor;
        else if (preferedVisual == @symbol(PseudoColor)) 
            cls = PseudoColor;
        else if (preferedVisual == @symbol(TrueColor)) 
            cls = TrueColor;
        else if (preferedVisual == @symbol(DirectColor)) 
            cls = DirectColor;
        else
            cls = PseudoColor;
        
        if (XMatchVisualInfo(dpy, screen, depth, cls, &vi)) {
            visual.visualid = vi.visualid;
printf("visualId=%x\n", vi.visualid);
        }
    }

    BLOCKINTERRUPTS();
    newWindow = XCreateWindow(dpy, parentWindow,
                           sizehints.x, sizehints.y,
                           sizehints.width, sizehints.height,
                           bw, depth, ioClass, &visual,
                           flags, &xswa);
    UNBLOCKINTERRUPTS();

    if (! newWindow) {
        RETURN ( nil );
    }

    /*
     * define its cursor
     */
    if (_isSmallInteger(wcursorId)) {
        XDefineCursor(dpy, newWindow, _CursorVal(wcursorId));
    }

    /*
     * define its icon and name
     * (only makes sense for topWindows)
     */
    if (isTopWindow) {
        if (_isSmallInteger(wiconId))
            iconBitmap = _PixmapVal(wiconId);
        else
            iconBitmap = (Pixmap)0;

        if (_isSmallInteger(wiconViewId))
            iconWindow = _WindowVal(wiconViewId);
        else
            iconWindow = (Window)0;

        if (__isString(wlabel) || __isSymbol(wlabel))
            windowName = (char *)_stringVal(wlabel);
        else
            windowName = "";

        if (iconBitmap || windowName) {
            XSetStandardProperties(dpy, newWindow,
                                        windowName, windowName,
                                        iconBitmap,
                                        0, 0, &sizehints);

        }

        wmhints.flags = 0;
        if (iconBitmap) {
            wmhints.flags |= IconPixmapHint;
            wmhints.icon_pixmap = iconBitmap;
        }
        if (iconWindow) {
            wmhints.flags |= IconWindowHint;
            wmhints.icon_window = iconWindow;
        }
/*
        wmhints.flags |= InputHint;
        wmhints.input = True;
*/
        XSetWMHints(dpy, newWindow, &wmhints);

        /*
         * tell window manager to not kill us but send an event instead
         */
        /*
         * get atoms first (if not already known)
         */
        if (_INST(ProtocolsAtom) == nil) {
            WmProtocolsAtom = XInternAtom(dpy, "WM_PROTOCOLS", False);
            _INST(ProtocolsAtom) = _MKSMALLINT(WmProtocolsAtom);
#ifdef USE_SAVEYOURSELF_ATOM
            WmSaveYourselfAtom = XInternAtom(dpy, "WM_SAVE_YOURSELF", False);
            _INST(SaveYourselfAtom) = _MKSMALLINT(WmSaveYourselfAtom);
#endif
#ifdef USE_QUIT_APP_ATOM
            WmQuitAppAtom = XInternAtom(dpy, "_WM_QUIT_APP", False);
            _INST(QuitAppAtom) = _MKSMALLINT(WmQuitAppAtom);
#endif
            WmDeleteWindowAtom = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
            _INST(DeleteWindowAtom) = _MKSMALLINT(WmDeleteWindowAtom);
        } else {
#ifdef USE_QUIT_APP_ATOM
            WmQuitAppAtom = _intVal(_INST(QuitAppAtom));
#else
            WmQuitAppAtom = 0;
#endif
            WmProtocolsAtom = _intVal(_INST(ProtocolsAtom));
            WmDeleteWindowAtom = _intVal(_INST(DeleteWindowAtom));
#ifdef USE_SAVEYOURSELF_ATOM
            WmSaveYourselfAtom = _intVal(_INST(SaveYourselfAtom));
#else
            WmSaveYourselfAtom = 0;
#endif
        }

        atoms[0] = WmDeleteWindowAtom; atomCount++;
#ifdef USE_SAVEYOURSELF_ATOM
        atoms[atomCount] = WmSaveYourselfAtom; atomCount++;
#endif
#ifdef USE_QUIT_APP_ATOM
        atoms[atomCount] = WmQuitAppAtom; atomCount++;
#endif
        XChangeProperty(dpy, newWindow, WmProtocolsAtom, XA_ATOM,
                        32, PropModeReplace, (unsigned char *)atoms, atomCount);
    }

    windowId = MKOBJ(newWindow);
%}
.
    self addKnownView:aView withId:windowId.
    ^ windowId
!

destroyView:aView withId:aWindowId
%{
    if (_isSmallInteger(aWindowId)) {
        XDestroyWindow(myDpy, _WindowVal(aWindowId));
    }
%}
.
    self removeKnownView:aView
!

destroyGC:aGCId

%{  /* NOCONTEXT */

    if (_isSmallInteger(aGCId)) {
        XFreeGC(myDpy, _GCVal(aGCId));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!


gcFor:aDrawableId

%{  /* NOCONTEXT */

    int screen = _intVal(_INST(screen));
    GC gc;

    if (_isSmallInteger(aDrawableId)) {
        gc = XCreateGC(myDpy, (Drawable)_WindowVal(aDrawableId),
                              0L, (XGCValues *)0);
        RETURN ( gc ? MKOBJ(gc) : nil );
    }
%}
.
    self primitiveFailed.
    ^ nil
!

dpsContextFor:aDrawableId and:aGCId

%{  /* NOCONTEXT */
#ifdef XXDPS
    int screen = _intVal(_INST(screen));
    DPSContext dps;
    int height;

    if (_isSmallInteger(aDrawableId) && _isSmallInteger(aGCId)) {
        dps = XDPSCreateContext(myDpy, (Drawable)_WindowVal(aDrawableId),
                                       _GCVal(aGCId),
                                       0, height, 0, colormap, NULL, 0,
                                       XDPSDefaultTextBackstop,
                                       XDPSDefaultErrorProc,
                                       NULL);
        RETURN ( dps ? MKOBJ(dps) : nil );
    }
#endif
%}
.
    self primitiveFailed.
    ^ nil
! !

!XWorkstation methodsFor:'resources'!

getResource:name class:cls
%{
    char *rslt;

    if ((__isString(name) || __isSymbol(name))
     && (__isString(cls) || __isSymbol(cls))) {
        rslt = XGetDefault(myDpy, (char *)_stringVal(cls),
                                  (char *)_stringVal(name));
        RETURN (rslt ? _MKSTRING(rslt,  __context) : nil );
    }
%}
.
    self primitiveFailed.
    ^ nil
! !

!XWorkstation methodsFor:'font stuff'!

decomposeXFontName:aString into:aBlock
    "extract family, face, style and size from an
     X-font name 
     (-brand-family-face-style-moreStyle--height-size-res-res-?-??-coding)
     ; evaluate aBlock with these"

    |origin family face style moreStyle skip height size
     resX resY x1 x2 coding start end |

    aString isNil ifTrue:[^ false].
    (aString startsWith:'-') ifFalse:[^ false].

    end := aString indexOf:$- startingAt:2.
    (end == 0) ifTrue:[^ false].
    origin := aString copyFrom:2 to:(end - 1).

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    family := aString copyFrom:start to:(end - 1).

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    face := aString copyFrom:start to:(end - 1).

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    style := aString copyFrom:start to:(end - 1).
    (style = 'o') ifTrue:[
        style := 'oblique'
    ] ifFalse:[
        (style = 'i') ifTrue:[
            style := 'italic'
        ] ifFalse:[
            (style = 'r') ifTrue:[
                style := 'roman'
            ]
        ]
    ].

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    moreStyle := aString copyFrom:start to:(end - 1).

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    skip := aString copyFrom:start to:(end - 1).

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    height := aString copyFrom:start to:(end - 1).

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    size := aString copyFrom:start to:(end - 1).
    size := (Number readFromString:size) / 10.

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    resX := aString copyFrom:start to:(end - 1).

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    resY := aString copyFrom:start to:(end - 1).

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    x1 := aString copyFrom:start to:(end - 1).

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    x2 := aString copyFrom:start to:(end - 1).

    start := end + 1.
    end := aString indexOf:$- startingAt:start.
    (end == 0) ifTrue:[^ false].
    coding := aString copyFrom:start to:(end - 1).

    aBlock value:family value:face value:style value:size value:coding
!

listOfAvailableFonts
    "return a list with all available font names on this display.
     Since this takes a long time, keep the result of the query for the
     next time."

    |stream aName arr|

    listOfXFonts isNil ifTrue:[
        stream := PipeStream readingFrom:'xlsfonts ''*'''.
        stream isNil ifTrue:[^ nil].
        listOfXFonts := VariableArray new.
        [stream atEnd] whileFalse:[
            aName := stream nextLine.
            aName notNil ifTrue:[
                self decomposeXFontName:aName into:
                    [:family :face :style :size :coding |
                        arr := Array with:family asSymbol
                                     with:face   asSymbol
                                     with:style  asSymbol
                                     with:size
                                     with:coding asSymbol.
                        listOfXFonts add:arr
                    ]
            ]
        ].
        stream close.
        "if xlsfont is broken ... (hey sco)"
        (listOfXFonts size == 0) ifTrue:[
            listOfXFonts := nil
        ]
    ].
    ^ listOfXFonts

    "Display listOfAvailableFonts"
!

getFontWithFamily:familyString face:faceString
            style:styleString size:sizeArg encoding:encodingSym

    "try to get the specified font, if not available, try next smaller
     font"

    |theSize theName theId xlatedStyle|

    "special: if face is nil, allow access to X-fonts"
    faceString isNil ifTrue:[
        theId := self createFontFor:familyString.
        theId isNil ifTrue:[
            theId := self getDefaultFont
        ].
        ^ theId
    ].

    xlatedStyle := styleString.
    "oblique is named italic in times font"
    ((familyString = 'Times') or:[familyString = 'times']) ifTrue:[
        ((styleString = 'Oblique') or:[styleString = 'oblique']) ifTrue:[
            xlatedStyle := 'italic'
        ]
    ].
    (xlatedStyle = 'italic') ifTrue:[
        xlatedStyle := 'i'
    ] ifFalse:[
        (xlatedStyle = 'roman') ifTrue:[
            xlatedStyle := 'r'
        ] ifFalse:[
            (xlatedStyle = 'oblique') ifTrue:[
                xlatedStyle := 'o'
            ]
        ]
    ].

    theId := nil.
    theSize := sizeArg.
    [theId isNil] whileTrue:[
        "this works only on Release >= 3 - X-servers"
        theName := ('-*-' , familyString ,
                     '-' , faceString ,
                     '-' , xlatedStyle , '-*-*-*-'
                     , theSize printString , '0-*-*-*-*-'
                     , encodingSym , '-1').
"
Transcript showCr:theName; endEntry.
"
        theId := self createFontFor:theName.
        theId isNil ifTrue:[
            "could not get the font - try next smaller one"
            theSize := theSize - 1.
            (theSize < (sizeArg // 2)) ifTrue:[
                "thats too much - give up"
                ^ self getDefaultFont
                "^ nil"
            ]
        ]
    ].
    (theSize ~~ sizeArg) ifTrue:[
        Transcript show:'next smaller font: '.
        Transcript showCr:theName
    ].
    ^ theId
!

getFontWithFoundry:foundry family:family weight:weight
              slant:slant size:size registry:registry encoding:encoding

    "get the specified font, if not available, return nil.
     This is the new font creation method - all others will be changed to
     use this wntry."

    |theName|

    "this works only on 'Release >= 3' - X-servers"
    theName := ('-' , foundry ,
                '-' , family ,
                '-' , weight ,
                '-' , slant , 
                '-*-*-*' ,
                '-' , size printString , '0' ,
                '-*-*-*-*' ,
                '-' , registry ,
                '-' , encoding).
Transcript showCr:theName; endEntry.

    ^ self createFontFor:theName.

    "
     Display getFontWithFoundry:'*'
                         family:'courier'
                         weight:'medium'
                          slant:'r'
                           size:13
                       registry:'iso8859'
                       encoding:'*'
    "
!

createFontFor:aFontName
    "a basic method for X-font allocation; this method allows
     any font to be aquired (even those not conforming to
     standard naming conventions, such as cursor, fixed or k14)"

%{  /* NOCONTEXT */

    XFontStruct *newFont;

    if (__isString(aFontName) || __isSymbol(aFontName)) {
        newFont = XLoadQueryFont(myDpy, (char *)_stringVal(aFontName));
        RETURN ( newFont ? MKOBJ(newFont) : nil );
    }
%}
.
    self primitiveFailed.
    ^ nil
!

getDefaultFont
    "return a default font id - used when class Font cannot
     find anything usable"

     ^ self createFontFor:'fixed'
!

releaseFont:aFontId

%{  /* NOCONTEXT */

    XFontStruct *f;

    if (_isSmallInteger(aFontId)) {
        f = _FontVal(aFontId);
        XFreeFont(myDpy, f);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

ascentOf:aFontId

%{  /* NOCONTEXT */

    XFontStruct *f;

    if (_isSmallInteger(aFontId)) {
        f = _FontVal(aFontId);
        RETURN ( _MKSMALLINT(f->ascent) );
    }
%}
.
    self primitiveFailed.
    ^ nil
!

descentOf:aFontId

%{  /* NOCONTEXT */

    XFontStruct *f;

    if (_isSmallInteger(aFontId)) {
        f = _FontVal(aFontId);
        RETURN ( _MKSMALLINT(f->descent) );
    }
%}
.
    self primitiveFailed.
    ^ nil
!

minWidthOfFont:aFontId

%{  /* NOCONTEXT */

    XFontStruct *f;

    if (_isSmallInteger(aFontId)) {
        f = _FontVal(aFontId);
        RETURN ( _MKSMALLINT(f->min_bounds.width) );
    }
%}
.
    self primitiveFailed.
    ^ nil
!

maxWidthOfFont:aFontId

%{  /* NOCONTEXT */

    XFontStruct *f;

    if (_isSmallInteger(aFontId)) {
        f = _FontVal(aFontId);
        RETURN ( _MKSMALLINT(f->max_bounds.width) );
    }
%}
.
    self primitiveFailed.
    ^ nil
!

widthOf:aString inFont:aFontId

%{  /* NOCONTEXT */

    XFontStruct *f;
    char *cp;
    int len, n;
#ifdef TWOBYTESTRINGS
    extern OBJ TwoByteString;
#endif

    if (_isSmallInteger(aFontId)) {
        f = _FontVal(aFontId);
        if (__isString(aString) || __isSymbol(aString)) {
            n = _stringSize(aString);
            cp = (char *)_stringVal(aString);
            len = XTextWidth(f, cp, n);
            RETURN ( _MKSMALLINT(len) );
        }
#ifdef TWOBYTESTRINGS
        if (_Class(aString) == TwoByteString) {
            n = _byteArraySize(aString) / 2;
            cp = (char *) _stringVal(aString);
            len = XTextWidth16(f, (XChar2b *)cp, n);
            RETURN ( _MKSMALLINT(len) );
        }
#endif
    }
%}
.
    self primitiveFailed.
    ^ nil
!

widthOf:aString from:index1 to:index2 inFont:aFontId

%{  /* NOCONTEXT */

    XFontStruct *f;
    char *cp;
    int len, n, i1, i2;
#ifdef TWOBYTESTRINGS
    extern OBJ TwoByteString;
#endif

    if (_isSmallInteger(index1)
     && _isSmallInteger(index2)
     && _isSmallInteger(aFontId)) {
        f = _FontVal(aFontId);
        i1 = _intVal(index1) - 1;
        i2 = _intVal(index2) - 1;
        if (__isString(aString) || __isSymbol(aString)) {
            cp = (char *) _stringVal(aString);
            n = _stringSize(aString);
            if ((i1 >= 0) && (i2 >= i1) && (i2 < n)) {
                cp += i1;
                len = XTextWidth(f, cp, i2 - i1 + 1);
                RETURN ( _MKSMALLINT(len) );
            }
        }
#ifdef TWOBYTESTRINGS
        if (_Class(aString) == TwoByteString) {
            cp = (char *) _stringVal(aString);
            n = _byteArraySize(aString) / 2;
            if ((i1 >= 0) && (i2 >= i1) && (i2 < n)) {
                cp += (i1 * 2);
                len = XTextWidth16(f, (XChar2b *)cp, i2 - i1 + 1);
                RETURN ( _MKSMALLINT(len) );
            }
        }
#endif
    }
%}
.
    self primitiveFailed.
    ^ nil
!

sizesInFamily:aFamilyName face:aFaceName style:aStyleName
    "return a set of all available font sizes in aFamily/aFace/aStyle
     on this display.
     Redefined to handle X's special case of 0-size (which stands for any)"

    |sizes|

    sizes := super sizesInFamily:aFamilyName face:aFaceName style:aStyleName.
    (sizes includes:0) ifTrue:[
        "special: in X11R5 and above, size 0 means:
         there are scaled versions in all sizes available"

        ^ #(4 5 6 7 8 9 10 11 12 14 16 18 20 22 24 28 32 48 64)
    ].
    ^ sizes

    "
     Display sizesInFamily:'courier' face:'bold' style:'roman'
    "
! !

!XWorkstation methodsFor:'cursor stuff'!

destroyCursor:aCursorId

%{  /* NOCONTEXT */

    if (_isSmallInteger(aCursorId)) {
        XFreeCursor(myDpy, _CursorVal(aCursorId));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

createCursorSourceForm:sourceForm maskForm:maskForm hotX:hx hotY:hy
    "create a cursor given 2 bitmaps (source, mask) and a hotspot"

    |id sourceId maskId|

    sourceId := sourceForm id.
    maskId := maskForm id.
%{
    Cursor newCursor;
    XColor fgColor, bgColor;

    if (_isSmallInteger(sourceId)
     && _isSmallInteger(maskId)
     && _isSmallInteger(hx)
     && _isSmallInteger(hy)) {
        fgColor.red = 0;        /* fg is black */
        fgColor.green = 0;
        fgColor.blue = 0;
        bgColor.red = 0xFFFF;   /* bg is white */
        bgColor.green = 0xFFFF;
        bgColor.blue = 0xFFFF;

        newCursor = XCreatePixmapCursor(myDpy,
                                _PixmapVal(sourceId),
                                _PixmapVal(maskId),
                                &fgColor, &bgColor, _intVal(hx), _intVal(hy));
        if (newCursor != (Cursor)0)
            id = MKOBJ(newCursor);
    }
%}
.
    ^ id
!

shapeNumberFromSymbol:shape
    "given a shape-symbol, return the corresponding cursor-number"

    "this is pure X-knowlegde - but you may easily add more"

    (shape == #upLeftArrow)     ifTrue:[ ^ 132 ].
    (shape == #upRightHand)     ifTrue:[ ^  58 ].
    (shape == #upDownArrow)     ifTrue:[ ^  116 ].
    (shape == #leftRightArrow)  ifTrue:[ ^  108 ].
    (shape == #upLimitArrow)    ifTrue:[ ^  138 ].
    (shape == #downLimitArrow)  ifTrue:[ ^  16 ].
    (shape == #leftLimitArrow)  ifTrue:[ ^  70 ].
    (shape == #rightLimitArrow) ifTrue:[ ^  96 ].
    (shape == #text)            ifTrue:[ ^  152 ].
    (shape == #upRightArrow)    ifTrue:[ ^  44 ].
    (shape == #leftHand)        ifTrue:[ ^  60 ].
    (shape == #questionMark)    ifTrue:[ ^  92 ].
    (shape == #cross)           ifTrue:[ ^  0 ].
    (shape == #wait)            ifTrue:[ ^  150 ].
    (shape == #crossHair)       ifTrue:[ ^  130 ].
    (shape == #origin)          ifTrue:[ ^  144 ].
    (shape == #corner)          ifTrue:[ ^  78 ].
    (shape == #square)          ifTrue:[ ^  40 ].
    (shape == #fourWay)         ifTrue:[ ^  52 ].
    (shape == #crossCursor)     ifTrue:[ ^  0 ].
    ('invalid cursorShape:' , shape printString) errorPrintNewline.
    ^  0
!

createCursorShape:aShape
    "create a cursor given a shape-symbol"

    |number id|

    number := self shapeNumberFromSymbol:aShape.
%{
    Cursor newCursor;

    if (_isSmallInteger(number)) {
        newCursor = XCreateFontCursor(myDpy, _intVal(number));
        if (newCursor != (Cursor)0) 
            id = MKOBJ(newCursor);
    }
%}
.
    ^ id
!

colorCursor:aCursorId foreground:fgColor background:bgColor
    "change a cursors colors"

    |fgR fgG fgB bgR bgG bgB|

    fgR := fgColor red.
    fgG := fgColor green.
    fgB := fgColor blue.
    bgR := bgColor red.
    bgG := bgColor green.
    bgB := bgColor blue.

    fgR := self percentToXColorValue:fgR.
    fgG := self percentToXColorValue:fgG.
    fgB := self percentToXColorValue:fgB.
    bgR := self percentToXColorValue:bgR.
    bgG := self percentToXColorValue:bgG.
    bgB := self percentToXColorValue:bgB.

%{
    XColor fgcolor, bgcolor;

    if (_isSmallInteger(aCursorId)
     && _isSmallInteger(fgR) && _isSmallInteger(fgG) && _isSmallInteger(fgB)
     && _isSmallInteger(bgR) && _isSmallInteger(bgG) && _isSmallInteger(bgB)) {

        fgcolor.red = _intVal(fgR);
        fgcolor.green= _intVal(fgG);
        fgcolor.blue = _intVal(fgB);
        bgcolor.red = _intVal(bgR);
        bgcolor.green= _intVal(bgG);
        bgcolor.blue = _intVal(bgB);
        XRecolorCursor(myDpy, _CursorVal(aCursorId), &fgcolor, &bgcolor);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

grabKeyboardIn:aWindowId
    "grab the keyboard"

%{  /* NOCONTEXT */
    int result;

    if (_isSmallInteger(aWindowId)) {
        result = XGrabKeyboard(myDpy,
                               _WindowVal(aWindowId),
                               True /* False */,
                               GrabModeAsync,
                               GrabModeAsync,
                               CurrentTime);
        switch(result) {
            case AlreadyGrabbed: 
                printf("AlreadyGrabbed\n");
                RETURN (false);
            case GrabNotViewable: 
                printf("GrabNotViewable\n");
                RETURN (false);
            case GrabInvalidTime: 
                printf("InvalidTime\n");
                RETURN (false);
            case GrabFrozen: 
                printf("Frozen\n");
                RETURN (false);
        }

        RETURN ( true );
    }
%}
.
    self primitiveFailed
!

ungrabKeyboard
    "release the keyboard"

%{  /* NOCONTEXT */

    XUngrabKeyboard(myDpy, CurrentTime);
    XSync(myDpy, 0);
%}
!

grabPointerIn:aWindowId
    "grap the pointer - return true if ok"

    ^ self grabPointerIn:aWindowId withCursor:nil pointerMode:#async keyboardMode:#sync confineTo:nil
!

grabPointerIn:aWindowId withCursor:aCursorId pointerMode:pMode keyboardMode:kMode confineTo:confineId
    "grap the pointer - return true if ok"

%{  /* NOCONTEXT */

    int result;
    Window confineWin;
    Cursor curs;
    int pointer_mode, keyboard_mode;

    if (_isSmallInteger(aWindowId)) {
        if (_isSmallInteger(confineId)) 
            confineWin = _WindowVal(confineId);
        else
            confineWin = (Window) None;

        if (_isSmallInteger(aCursorId)) 
            curs = _CursorVal(aCursorId);
        else
            curs = (Cursor) None;

        if (pMode == @symbol(sync))
            pointer_mode = GrabModeSync;
        else
            pointer_mode = GrabModeAsync;

        if (kMode == @symbol(sync))
            keyboard_mode = GrabModeSync;
        else
            keyboard_mode = GrabModeAsync;

        result = XGrabPointer(myDpy,
                              _WindowVal(aWindowId), 
                              False, 
                              ButtonPressMask | ButtonMotionMask | ButtonReleaseMask,
                              pointer_mode, keyboard_mode,
                              confineWin,
                              curs,
                              CurrentTime);

        switch (result) {
            case AlreadyGrabbed: 
                printf("AlreadyGrabbed\n");
                RETURN (false);
            case GrabNotViewable: 
                printf("GrabNotViewable\n");
                RETURN (false);
            case GrabInvalidTime: 
                printf("InvalidTime\n");
                RETURN (false);
            case GrabFrozen: 
                printf("Frozen\n");
                RETURN (false);
        }

        RETURN ( true );
    }
%}
.
    self primitiveFailed
!

ungrabPointer
    "release the pointer"

%{  /* NOCONTEXT */

    XUngrabPointer(myDpy, CurrentTime);
    XSync(myDpy, 0);
%}
!

rootPositionOfLastEvent
    "return the position in root-window coordinates
     of the last button, key or pointer event"

    ^ eventRootX @ eventRootY
!

pointerPosition
    "return the current pointer position in root-window coordinates"

    |xpos ypos|
%{
    Display *dpy = myDpy;
    Window w;
    int screen = _intVal(_INST(screen));
    Window rootRet, childRet;
    int rootX, rootY, winX, winY;
    unsigned int mask;

#ifdef OLD
    w = RootWindow(dpy, screen);
#else
    w = getRootWindow(myDpy, screen);
#endif
    XQueryPointer(dpy, w, &rootRet, &childRet,
                          &rootX, &rootY,
                          &winX, &winY,
                          &mask);
    xpos = _MKSMALLINT(rootX);
    ypos = _MKSMALLINT(rootY);
%}
.
    ^ xpos @ ypos
!

buttonStates
    "return an integer representing the state of the pointer buttons;
     a one-bit in positions 0.. represent a pressed button"

%{  /* NOCONTEXT*/
    Display *dpy = myDpy;
    Window w;
    int screen = _intVal(_INST(screen));
    Window rootRet, childRet;
    int rootX, rootY, winX, winY;
    unsigned int mask;

#ifdef OLD
    w = RootWindow(dpy, screen);
#else
    w = getRootWindow(myDpy, screen);
#endif
    XQueryPointer(dpy, w, &rootRet, &childRet,
                          &rootX, &rootY,
                          &winX, &winY,
                          &mask);
    RETURN (_MKSMALLINT(mask));
%}
!

leftButtonPressed
    "return true, if the left button is currently pressed"

    ^ (self buttonStates bitAnd:256) ~~ 0
!

middleButtonPressed
    "return true, if the middle button is currently pressed.
     Notice, that there are systems with only a single button."

    ^ (self buttonStates bitAnd:512) ~~ 0
!

rightButtonPressed
    "return true, if the left button is currently pressed.
     Notice, that there are systems with only one or two buttons."

    ^ (self buttonStates bitAnd:1024) ~~ 0
! !

!XWorkstation methodsFor:'color stuff'!

listOfAvailableColors
    "return a list of all available colornames.
     This should not be used, since colornames are very
     display-specific (here X-specific)."

    |aStream list line index colorName|

    aStream := FileStream readonlyFileNamed:'/usr/lib/X11/rgb.txt'.
    aStream isNil ifTrue:[^ nil].
    list := Text new.
    [aStream atEnd] whileFalse:[
        line := aStream nextLine.
        line notNil ifTrue:[
            "skip the r/g/b numbers"
            index := 1.
            [(line at:index) isDigit] whileTrue:[index := index + 1].
            [(line at:index) isSeparator] whileTrue:[index := index + 1].
            [(line at:index) isDigit] whileTrue:[index := index + 1].
            [(line at:index) isSeparator] whileTrue:[index := index + 1].
            [(line at:index) isDigit] whileTrue:[index := index + 1].
            [(line at:index) isSeparator] whileTrue:[index := index + 1].
            colorName := line copyFrom:index.
            ((colorName occurrencesOf:(Character space)) == 0) ifTrue:[
                list add:colorName
            ]
        ]
    ].
    aStream close.
    ^ list sort
!

freeColor:colorIndex
    "free a display color when its no longer needed"

%{  /* NOCONTEXT */

    Display *dpy = myDpy;
    unsigned long color;
    int screen = _intVal(_INST(screen));

    if (_isSmallInteger(colorIndex)) {
        color = (long) _intVal(colorIndex);
        XFreeColors(dpy, DefaultColormap(dpy, screen), &color, 1, 0L);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

percentToXColorValue:aPercentage
    "given a color-component value in percent (0..100), return the corresponding
     x-component value (0..65k) as an integer"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aPercentage)) {
        RETURN ( _MKSMALLINT(0xFFFF * _intVal(aPercentage) / 100) );
    }
    if (__isFloat(aPercentage)) {
        RETURN ( _MKSMALLINT(0xFFFF * (int)(_floatVal(aPercentage)) / 100) );
    }
%}
.
    ^ (16rFFFF * aPercentage / 100) rounded
!

colorRed:redVal green:greenVal blue:blueVal
    "allocate a color with rgb values (0..100) - return index"

    |r g b|

    r := self percentToXColorValue:redVal.
    g := self percentToXColorValue:greenVal.
    b := self percentToXColorValue:blueVal.
%{
    Display *dpy = myDpy;
    XColor ecolor;
    int screen = _intVal(_INST(screen));

    if (_isSmallInteger(r)
     && _isSmallInteger(g)
     && _isSmallInteger(b)) {
        ecolor.red = _intVal(r);
        ecolor.green= _intVal(g);
        ecolor.blue = _intVal(b);
        if (!XAllocColor(dpy, DefaultColormap(dpy, screen), &ecolor)) {
            RETURN ( nil );
        }
        RETURN ( _MKSMALLINT(ecolor.pixel) );
    }
%}
.
    self primitiveFailed.
    ^ nil
!

colorNamed:aString
    "allocate a color with color name - return index.
     Dont use this method, colornames are mostly X specific"

%{  /* NOCONTEXT */

    char *colorname;
    Display *dpy = myDpy;
    XColor scolor, ecolor;
    int screen =_intVal(_INST(screen));

    if (__isString(aString) || __isSymbol(aString)) {
        colorname = (char *)_stringVal(aString);

        if (!XParseColor(dpy, DefaultColormap(dpy, screen),
                              colorname, &ecolor)) {
            RETURN ( nil );
        }
        if (!XAllocColor(dpy, DefaultColormap(dpy, screen), &ecolor)) {
            RETURN ( nil );
        }
        RETURN ( _MKSMALLINT(ecolor.pixel) );
    }
%}
.
    self primitiveFailed.
    ^ nil
!

colorCell
    "allocate a color cell - return index.
     This method will return nil for StaticGrey and StaticGrey displays."

%{  /* NOCONTEXT */

    Display *dpy = myDpy;
    int screen = _intVal(_INST(screen));
    XColor color;
    unsigned long dummy;
    Status ok;

    ok = XAllocColorCells(dpy, DefaultColormap(dpy, screen), (Bool)0,
                               &dummy, 0, &color.pixel, 1);
    if (ok) {
        RETURN ( _MKSMALLINT(color.pixel) );
    }
%}
.
    ^ nil
!

setColor:index red:redVal green:greenVal blue:blueVal
    "change color in map at:index to rgb (0..100).
     This method is a noop for StaticGrey and StaticGrey displays."

%{  /* NOCONTEXT */

    char *colorname;
    Display *dpy = myDpy;
    XColor color;
    int screen = _intVal(_INST(screen));
    int r, g, b;
    int ok = 1;

    if (_isSmallInteger(redVal))
        r = 0xFFFF * _intVal(redVal) / 100;
    else if (__isFloat(redVal))
        r = 0xFFFF * (int)(_floatVal(redVal)) / 100;
    else ok = 0;

    if (_isSmallInteger(greenVal))
        g = 0xFFFF * _intVal(greenVal) / 100;
    else if (__isFloat(greenVal))
        g = 0xFFFF * (int)(_floatVal(greenVal)) / 100;
    else ok = 0;

    if (_isSmallInteger(blueVal))
        b = 0xFFFF * _intVal(blueVal) / 100;
    else if (__isFloat(blueVal))
        b = 0xFFFF * (int)(_floatVal(blueVal)) / 100;
    else ok = 0;

    if (_isSmallInteger(index) && ok) {
        color.pixel = _intVal(index);
        color.red = r;
        color.green = g;
        color.blue = b;
        color.flags = DoRed | DoGreen | DoBlue;

        XStoreColor(dpy, DefaultColormap(dpy, screen), &color);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

getRGBFromName:colorName into:aBlock
    "get rgb components (0..100) of color named colorName,
     and evaluate the 3-arg block, aBlock with them"

    |r g b|
%{
    Display *dpy = myDpy;
    int screen = _intVal(_INST(screen));
    XColor color;
    double fr, fg, fb;
    double floor();

    if (__isString(colorName) || __isSymbol(colorName)) {
        if (XParseColor(dpy, DefaultColormap(dpy, screen), (char *)_stringVal(colorName),
                                                           &color)) {
            /* scale to 0..100 and round to the first decimal */
            fr = floor( ((((double)color.red) * 1000.0) / 0xFFFF) + 0.5) / 10.0;
            fg = floor( ((((double)color.green) * 1000.0) / 0xFFFF) + 0.5) / 10.0;
            fb = floor( ((((double)color.blue) * 1000.0) / 0xFFFF) + 0.5) / 10.0;

            r = _MKFLOAT(fr, __context);
            g = _MKFLOAT(fg, __context);
            b = _MKFLOAT(fb, __context);
        }
    }
%}
.
    aBlock value:r value:g value:b
!

getRGBFrom:index into:aBlock
    "get rgb components (0..100) of color in map at:index,
     and evaluate the 3-arg block, aBlock with them"

    |r g b|
%{
    Display *dpy = myDpy;
    int screen = _intVal(_INST(screen));
    XColor color;
    double fr, fg, fb;
    double floor();
    int bits, scale, shift;

    if (_isSmallInteger(index)) {
        color.pixel = _intVal(index);
        XQueryColor(dpy, DefaultColormap(dpy, screen), &color);

        /* scale to 0..100 and round to the first decimal */

        /* 
         * have to compensate for an error in X ?, which does not scale
         * colors correctly if lesser than 16bits are valid in a color,
         * (for example, color white on a 4bitsPerRGB server will return
         * (16rF000 16rF000 16rF000) instead of (16rFFFF 16rFFFF 16rFFFF)
         */
        bits = _intVal(_INST(bitsPerRGB));
        scale = (1<<bits) - 1;
        shift = 16 - bits;

        fr = floor( ( ((double)(color.red>>shift) * 1000.0) / scale) + 0.5) / 10.0;
        fg = floor( ( ((double)(color.green>>shift) * 1000.0) / scale) + 0.5) / 10.0;
        fb = floor( ( ((double)(color.blue>>shift) * 1000.0) / scale) + 0.5) / 10.0;

        r = _MKFLOAT(fr, __context);
        g = _MKFLOAT(fg, __context);
        b = _MKFLOAT(fb, __context);
    }
%}
.
    aBlock value:r value:g value:b
! !

!XWorkstation methodsFor:'window stuff'!

setBackingStore:how in:aWindowId
    "turn on/off backing-store for a window"

%{  /* NOCONTEXT */

    XSetWindowAttributes wa;

    if (_isSmallInteger(aWindowId)) {
        if (_INST(ignoreBackingStore) != true) {
            if (how == @symbol(always)) wa.backing_store = Always;
            else if (how == @symbol(whenMapped)) wa.backing_store = WhenMapped;
            else if (how == true) wa.backing_store = Always;
            else wa.backing_store = 0;
            XChangeWindowAttributes(myDpy, _WindowVal(aWindowId), CWBackingStore, &wa);
        }
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setSaveUnder:yesOrNo in:aWindowId
    "turn on/off save-under for a window"

%{  /* NOCONTEXT */

    XSetWindowAttributes wa;

    if (_isSmallInteger(aWindowId)) {
        if (_INST(hasSaveUnder) == true) {
            wa.save_under = (yesOrNo == true) ? 1 : 0;
            XChangeWindowAttributes(myDpy, _WindowVal(aWindowId), CWSaveUnder, &wa);
        }
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setWindowBackground:aColorIndex in:aWindowId
    "set the windows background color"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId) && _isSmallInteger(aColorIndex)) {
        XSetWindowBackground(myDpy, _WindowVal(aWindowId), _intVal(aColorIndex));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setWindowBackgroundPixmap:aPixmapId in:aWindowId
    "set the windows background pattern to be a form"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId) && _isSmallInteger(aPixmapId)) {
        XSetWindowBackgroundPixmap(myDpy, _WindowVal(aWindowId), _PixmapVal(aPixmapId));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setWindowBorderColor:aColorIndex in:aWindowId
    "set the windows border color"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId) && _isSmallInteger(aColorIndex)) {
        XSetWindowBorder(myDpy, _WindowVal(aWindowId), _intVal(aColorIndex));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setWindowBorderPixmap:aPixmapId in:aWindowId
    "set the windows border pattern"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId) && _isSmallInteger(aPixmapId)) {
        XSetWindowBorderPixmap(myDpy, _WindowVal(aWindowId), _PixmapVal(aPixmapId));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setWindowBorderWidth:aNumber in:aWindowId
    "set the windows border width"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId) && _isSmallInteger(aNumber)) {
        XSetWindowBorderWidth(myDpy, _WindowVal(aWindowId), _intVal(aNumber));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setWindowBorderShape:aPixmapId in:aWindowId
    "set the windows border shape"

    hasShapeExtension ifFalse:[^ self].

%{  /* NOCONTEXT */

#ifdef SHAPE
    if (_isSmallInteger(aWindowId) && _isSmallInteger(aPixmapId)) {
        XShapeCombineMask(myDpy, _WindowVal(aWindowId), ShapeBounding,
                          0, 0, _PixmapVal(aPixmapId), ShapeSet);
        RETURN ( self );
    }
#endif
%}
.
    self primitiveFailed
!

setWindowShape:aPixmapId in:aWindowId
    "set the windows shape"

    hasShapeExtension ifFalse:[^ self].

%{  /* NOCONTEXT */

#ifdef SHAPE
    if (_isSmallInteger(aWindowId) && _isSmallInteger(aPixmapId)) {
        XShapeCombineMask(myDpy, _WindowVal(aWindowId), ShapeClip,
                          0, 0,
                          _PixmapVal(aPixmapId), ShapeSet);
        RETURN ( self );
    }
#endif
%}
.
    self primitiveFailed
!

setCursor:aCursorId in:aWindowId
    "define a windows cursor"

%{  /* NOCONTEXT */

    Display *dpy = myDpy;

    if (_isSmallInteger(aWindowId) && _isSmallInteger(aCursorId)) {
        XDefineCursor(dpy, _WindowVal(aWindowId), _CursorVal(aCursorId));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setWindowName:aString in:aWindowId
    "define a windows name"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId)
     && (__isString(aString) || __isSymbol(aString))) {
        XStoreName(myDpy, _WindowVal(aWindowId), (char *)_stringVal(aString));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setIconName:aString in:aWindowId
    "define a windows iconname"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId)
     && (__isString(aString) || __isSymbol(aString))) {
        XSetIconName(myDpy, _WindowVal(aWindowId), (char *)_stringVal(aString));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setWindowIcon:aForm in:aWindowId
    "define a bitmap to be used as icon"

    |iconId|

    aForm notNil ifTrue:[
        iconId := aForm id
    ].
%{
    if (_isSmallInteger(iconId) && _isSmallInteger(aWindowId)) {
        XWMHints hints;

        hints.icon_pixmap = _PixmapVal(iconId);
        hints.flags = IconPixmapHint;
        XSetWMHints(myDpy, _WindowVal(aWindowId), &hints);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setWindowIconWindow:aView in:aWindowId
    "define a window to be used as icon"

    |iconWindowId|

    aView notNil ifTrue:[
        iconWindowId := aView id
    ].
%{
    if (_isSmallInteger(iconWindowId) && _isSmallInteger(aWindowId)) {
        XWMHints hints;

        hints.icon_window = _WindowVal(iconWindowId);
        hints.flags = IconWindowHint;
        XSetWMHints(myDpy, _WindowVal(aWindowId), &hints);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

clearWindow:aWindowId
    "clear a window to viewbackground"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId)) {
        XClearWindow(myDpy, _WindowVal(aWindowId));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

clearRectangleX:x y:y width:width height:height in:aWindowId
    "clear a rectangular area to viewbackground"

%{  /* NOCONTEXT */

    int w, h;

    if (_isSmallInteger(aWindowId)
     && _isSmallInteger(x) && _isSmallInteger(y)
     && _isSmallInteger(width) && _isSmallInteger(height)) {
        w = _intVal(width);
        h = _intVal(height);
        if (w < 0) w = 0;
        if (h < 0) h = 0;
        XClearArea(myDpy, _WindowVal(aWindowId), _intVal(x), _intVal(y), w, h, 0);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

mapView:aView id:aWindowId iconified:aBoolean atX:xPos y:yPos width:w height:h
    "make a window visible - either as icon or as a real view - needed for restart"

    |wicon wiconId wiconView wiconViewId wlabel|

    aBoolean ifTrue:[
        wicon := aView icon.
        wicon notNil ifTrue:[
            wiconId := wicon id
        ].
        wiconView := aView iconView.
        wiconView notNil ifTrue:[
            wiconViewId := wiconView id
        ].
        wlabel := aView label.
    ].
%{  

    XWMHints wmhints;
    XSizeHints szhints;
    Display *dpy = myDpy;
    Window win;

    if (_isSmallInteger(aWindowId)) {
        win = _WindowVal(aWindowId);

        szhints.flags = 0;
        if (_isSmallInteger(xPos) && _isSmallInteger(yPos)) {
            szhints.x = _intVal(xPos);
            szhints.y = _intVal(yPos);
            szhints.flags |= USPosition;
        }
        if (_isSmallInteger(w) && _isSmallInteger(h)) {
            szhints.width = _intVal(w);
            szhints.height = _intVal(h);
            szhints.flags |= USSize;
        }

        if (aBoolean == true) {
            char *windowName;
            Pixmap iconBitmap = (Pixmap)0;
            Window iconWindow;

            if (_isSmallInteger(wiconId))
                iconBitmap = _PixmapVal(wiconId);
            else
                iconBitmap = (Pixmap)0;

            if (_isSmallInteger(wiconViewId))
                iconWindow = _WindowVal(wiconViewId);
            else
                iconWindow = (Window)0;

            if (__isString(wlabel) || __isSymbol(wlabel))
                windowName = (char *)_stringVal(wlabel);
            else
                windowName = "";

            if (iconBitmap || windowName) {
                XSetStandardProperties(dpy, win,
                                        windowName, windowName,
                                        iconBitmap,
                                        0, 0, &szhints);
            }

            wmhints.flags = 0;
            if (iconBitmap) {
                wmhints.flags |= IconPixmapHint;
                wmhints.icon_pixmap = iconBitmap;
            }
            if (iconWindow) {
                wmhints.flags |= IconWindowHint;
                wmhints.icon_window = iconWindow;
            }

            wmhints.initial_state = IconicState;
            wmhints.flags |= StateHint;
            XSetWMHints(dpy, win, &wmhints);
        }

        if (szhints.flags) {
            XSetNormalHints(dpy, win, &szhints);
        }

        XMapWindow(dpy, win);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

mapWindow:aWindowId
    "make a window visible"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId)) {
        XMapWindow(myDpy, _WindowVal(aWindowId));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

unmapWindow:aWindowId
    "make a window invisible"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId)) {
        XUnmapWindow(myDpy, _WindowVal(aWindowId));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

raiseWindow:aWindowId
    "bring a window to front"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId)) {
        XRaiseWindow(myDpy, _WindowVal(aWindowId));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

lowerWindow:aWindowId
    "bring a window to back"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId)) {
        XLowerWindow(myDpy, _WindowVal(aWindowId));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

moveWindow:aWindowId x:x y:y
    "move a window"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aWindowId)
     && _isSmallInteger(x) && _isSmallInteger(y)) {
        XMoveWindow(myDpy, _WindowVal(aWindowId), _intVal(x), _intVal(y));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

resizeWindow:aWindowId width:w height:h
    "resize a window"

%{  /* NOCONTEXT */
    
    int newWidth, newHeight;

    if (_isSmallInteger(aWindowId)
     && _isSmallInteger(w) && _isSmallInteger(h)) {
        newWidth = _intVal(w);
        newHeight = _intVal(h);
        if (newWidth < 1) newWidth = 1;
        if (newHeight < 1) newHeight = 1;
        XResizeWindow(myDpy, _WindowVal(aWindowId), newWidth, newHeight);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

moveResizeWindow:aWindowId x:x y:y width:w height:h
    "move and resize a window"

%{  /* NOCONTEXT */
 
    int newWidth, newHeight;

    if (_isSmallInteger(aWindowId)
     && _isSmallInteger(w) && _isSmallInteger(h)
     && _isSmallInteger(x) && _isSmallInteger(y)) {
        newWidth = _intVal(w);
        newHeight = _intVal(h);
        if (newWidth < 1) newWidth = 1;
        if (newHeight < 1) newHeight = 1;
        XMoveResizeWindow(myDpy, _WindowVal(aWindowId),
                              _intVal(x), _intVal(y),
                              newWidth, newHeight);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
! !

!XWorkstation methodsFor:'graphic context stuff'!

setForeground:fgColorIndex in:aGCId
    "set foreground color to be drawn with"

%{  /* NOCONTEXT */

    if (_isSmallInteger(fgColorIndex) && _isSmallInteger(aGCId)) {
        XSetForeground(myDpy, _GCVal(aGCId), _intVal(fgColorIndex));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setBackground:bgColorIndex in:aGCId
    "set background color to be drawn with"

%{  /* NOCONTEXT */

    if (_isSmallInteger(bgColorIndex) && _isSmallInteger(aGCId)) {
        XSetBackground(myDpy, _GCVal(aGCId), _intVal(bgColorIndex));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setForeground:fgColorIndex background:bgColorIndex in:aGCId
    "set foreground and background colors to be drawn with"

%{  /* NOCONTEXT */

    Display *dpy = myDpy;
    GC gc = _GCVal(aGCId);

    if (_isSmallInteger(fgColorIndex) && _isSmallInteger(bgColorIndex)
     && _isSmallInteger(aGCId)) {
        XSetForeground(dpy, gc, _intVal(fgColorIndex));
        XSetBackground(dpy, gc, _intVal(bgColorIndex));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setForeground:fgColor background:bgColor mask:aBitmapId in:aGCId
    "set foreground and background colors to be drawn with using mask or
     solid (if aBitmapId is nil)"

%{  /* NOCONTEXT */

    Display *dpy = myDpy;
    GC gc = _GCVal(aGCId);

    if (_isSmallInteger(aGCId)) {
        if (_isSmallInteger(fgColor))
            XSetForeground(dpy, gc, _intVal(fgColor));
        if (_isSmallInteger(bgColor))
            XSetBackground(dpy, gc, _intVal(bgColor));

        if (_isSmallInteger(aBitmapId)) {
            XSetStipple(dpy, gc, _PixmapVal(aBitmapId));
            XSetFillStyle(dpy, gc, FillOpaqueStippled);
            RETURN ( self );
        }
        if (aBitmapId == nil) {
            XSetFillStyle(dpy, gc, FillSolid);
            RETURN ( self );
        }
    }
%}
.
    self primitiveFailed
!

setLineWidth:aNumber style:lineStyle cap:capStyle join:joinStyle in:aGCId
    "set line attributes"

%{  /* NOCONTEXT */

    int x_style, x_cap, x_join;

    if (_isSmallInteger(aGCId)) {
        if (_isSmallInteger(aNumber)) {
            if (lineStyle == @symbol(solid)) x_style = LineSolid;
            else if (lineStyle == @symbol(dashed)) x_style = LineOnOffDash;
            else if (lineStyle == @symbol(doubleDashed)) x_style = LineDoubleDash;
            else  goto bad;

            if (capStyle == @symbol(notLast)) x_cap = CapNotLast;
            else if (capStyle == @symbol(butt)) x_cap = CapButt;
            else if (capStyle == @symbol(round)) x_cap  = CapRound;
            else if (capStyle == @symbol(projecting)) x_cap  = CapProjecting;
            else  goto bad;

            if (joinStyle == @symbol(miter)) x_join = JoinMiter;
            else if (joinStyle == @symbol(bevel)) x_join = JoinBevel;
            else if (joinStyle == @symbol(round)) x_join  = JoinRound;
            else  goto bad;

            XSetLineAttributes(myDpy,
                           _GCVal(aGCId), _intVal(aNumber),
                           x_style, x_cap, x_join);
            RETURN ( self );
        }
    }
bad: ;
%}
.
    self primitiveFailed
!

setForeground:fgColor background:bgColor mask:aBitmapId lineWidth:lw in:aGCId
    "set foreground and background colors to be drawn with using mask or
     solid (if aBitmapId is nil); also set lineWidth"

%{  /* NOCONTEXT */

    Display *dpy = myDpy;
    GC gc = _GCVal(aGCId);

    if (_isSmallInteger(aGCId)) {
        if (_isSmallInteger(lw)) {
            XSetLineAttributes(dpy, gc, _intVal(lw),
                                    LineSolid, CapNotLast, JoinMiter);
        }
        if (_isSmallInteger(fgColor))
            XSetForeground(dpy, gc, _intVal(fgColor));
        if (_isSmallInteger(bgColor))
            XSetBackground(dpy, gc, _intVal(bgColor));

        if (_isSmallInteger(aBitmapId)) {
            XSetStipple(dpy, gc, _PixmapVal(aBitmapId));
            XSetFillStyle(dpy, gc, FillOpaqueStippled);
            RETURN ( self );
        }
        if (aBitmapId == nil) {
            XSetFillStyle(dpy, gc, FillSolid);
            RETURN ( self );
        }
    }
%}
.
    self primitiveFailed
!

setFunction:aFunctionSymbol in:aGCId
    "set alu function to be drawn with"

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    int fun = -1;

    if (_isSmallInteger(aGCId)) {
        if (aFunctionSymbol == @symbol(copy)) fun = GXcopy;
        else if (aFunctionSymbol == @symbol(copyInverted)) fun = GXcopyInverted;
        else if (aFunctionSymbol == @symbol(xor)) fun = GXxor;
        else if (aFunctionSymbol == @symbol(and)) fun = GXand;
        else if (aFunctionSymbol == @symbol(andReverse)) fun = GXandReverse;
        else if (aFunctionSymbol == @symbol(andInverted)) fun = GXandInverted;
        else if (aFunctionSymbol == @symbol(or)) fun = GXor;
        else if (aFunctionSymbol == @symbol(orReverse)) fun = GXorReverse;
        else if (aFunctionSymbol == @symbol(orInverted)) fun = GXorInverted;
        if (fun != -1) {
            XSetFunction(myDpy, gc, fun);
            RETURN ( self );
        }
    }
%}
.
    "
     either aGCId is not an integer, or an invalid symbol
     was passed ...
    "
    self primitiveFailed
!

setFont:aFontId in:aGCId
    "set font to be drawn in"

%{  /* NOCONTEXT */

    XFontStruct *f;

    if (_isSmallInteger(aFontId) && _isSmallInteger(aGCId)) {
        f = (XFontStruct *) _FontVal(aFontId);
        XSetFont(myDpy, _GCVal(aGCId), f->fid);
        RETURN ( self );
    }
%}
.
    "
     aGCId and/or aFontId are invalid
    "
    self primitiveFailed
!

setPixmapMask:aPixmapId in:aGCId
    "set or clear the drawing mask - a pixmap mask providing full color"

%{  /* NOCONTEXT */

    Display *dpy = myDpy;
    GC gc = _GCVal(aGCId);
    Pixmap pixmap;

    if (_isSmallInteger(aGCId)) {
        if (_isSmallInteger(aPixmapId)) {
            pixmap = _PixmapVal(aPixmapId);
            XSetTile(dpy, gc, pixmap);
            XSetFillStyle(dpy, gc, FillTiled);
            RETURN ( self );
        }
        if (aPixmapId == nil) {
            XSetFillStyle(dpy, gc, FillSolid);
            RETURN ( self );
        }
    }
%}
.
    self primitiveFailed
!

setBitmapMask:aBitmapId in:aGCId
    "set or clear the drawing mask - a bitmap mask using current fg/bg"

%{  /* NOCONTEXT */

    Display *dpy = myDpy;
    GC gc = _GCVal(aGCId);
    Pixmap bitmap;

    if (_isSmallInteger(aGCId)) {
        if (_isSmallInteger(aBitmapId)) {
            bitmap = _PixmapVal(aBitmapId);
            XSetStipple(dpy, gc, bitmap);
            XSetFillStyle(dpy, gc, FillOpaqueStippled);
            RETURN ( self );
        }
        if (aBitmapId == nil) {
            XSetFillStyle(dpy, gc, FillSolid);
            RETURN ( self );
        }
    }
%}
.
    self primitiveFailed
!

setMaskOriginX:orgX y:orgY in:aGCid
    "set the mask origin"

%{  /* NOCONTEXT */

    if (_isSmallInteger(orgX)
     && _isSmallInteger(orgY)
     && _isSmallInteger(aGCid)) {
        XSetTSOrigin(myDpy, _GCVal(aGCid), _intVal(orgX), _intVal(orgY));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setClipByChildren:aBool in:aGCId
    "enable/disable drawing into child views"

%{  /* NOCONTEXT */

    XGCValues gcv;
    GC gc = _GCVal(aGCId);

    if (_isSmallInteger(aGCId)) {
        if (aBool == true)
            gcv.subwindow_mode = ClipByChildren;
        else
            gcv.subwindow_mode = IncludeInferiors;

        XChangeGC(myDpy, gc, GCSubwindowMode, &gcv);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

noClipIn:aGCId
    "disable clipping rectangle"

%{  /* NOCONTEXT */

    XGCValues gcv;
    GC gc = _GCVal(aGCId);

    if (_isSmallInteger(aGCId)) {
        gcv.clip_mask = None;
        XChangeGC(myDpy, gc, GCClipMask, &gcv);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setClipX:clipX y:clipY width:clipWidth height:clipHeight in:aGCId
    "clip to a rectangle"

%{  /* NOCONTEXT */

    XRectangle r;

    if (_isSmallInteger(aGCId)
     && _isSmallInteger(clipX) && _isSmallInteger(clipY)
     && _isSmallInteger(clipWidth) && _isSmallInteger(clipHeight)) {
        r.x = _intVal(clipX);
        r.y = _intVal(clipY);
        r.width = _intVal(clipWidth);
        r.height = _intVal(clipHeight);
        XSetClipRectangles(myDpy, _GCVal(aGCId), 0, 0, &r, 1, Unsorted);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

setGraphicsExposures:aBoolean in:aGCId
    "set or clear the graphics exposures flag"

%{  /* NOCONTEXT */

    if (_isSmallInteger(aGCId)) {
        XSetGraphicsExposures(myDpy, _GCVal(aGCId), (aBoolean==true)?1:0);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
! !

!XWorkstation methodsFor:'retrieving pixels'!

getPixelX:x y:y from:aDrawableId
    "return the pixel value at x/y"
%{
    Window win = _WindowVal(aDrawableId);
    XImage *img;
    int ret;
    int xpos, ypos;

    if (_isSmallInteger(aDrawableId)
     && _isSmallInteger(x) && _isSmallInteger(y)) {
        xpos = _intVal(x);
        ypos = _intVal(y);
        if ((xpos < 0) || (ypos < 0)) {
            RETURN ( _MKSMALLINT(0) );
        }
        img = XGetImage(myDpy, win, xpos, ypos, 1, 1, (unsigned)~0, ZPixmap);
        ret = XGetPixel(img, 0, 0);
        XDestroyImage(img);
        RETURN (  _MKSMALLINT(ret) );
    }
%}
.
    self primitiveFailed.
    ^ nil
!

getBitsFrom:aDrawableId x:srcx y:srcy width:w height:h into:imageBits
    "get bits from a drawable into the imageBits. The storage for the bits
     be big enough for the data to fit. If ok, returns an array with some
     info and the bits in imageBits."

     |info|

     info := Array with:nil "depth"
                   with:nil "bit order"
                   with:nil "bytes_per_line".

%{
    {
        Display *dpy = myDpy;
        Window win = _WindowVal(aDrawableId);
        extern OBJ ByteArray;
        XImage *image;
        int pad, bytes_per_line, bytes;

        if (_isSmallInteger(aDrawableId)
         && _isSmallInteger(srcx) && _isSmallInteger(srcy)
         && _isSmallInteger(w) && _isSmallInteger(h)
         && __isArray(info)
         && __isByteArray(imageBits)) {
            image = XGetImage(dpy, win, _intVal(srcx), _intVal(srcy),
                                        _intVal(w), _intVal(h),
                                        (unsigned)AllPlanes, ZPixmap);

            pad = image->bitmap_pad;
#ifdef SUPERDEBUG
            printf("pad:%d depth:%d\n", image->bitmap_pad, image->depth);
#endif
            switch (image->depth) {
                case 1:
                case 2:
                case 4:
                case 8:
                case 16:
                case 24:
                    bytes = image->bytes_per_line * image->height;
                    break;
                default:
                    /* unsupported depth */
                    XDestroyImage(image);
                    goto fail;
            }

#ifdef SUPERDEBUG
            printf("bytes need:%d bytes given:%d\n", bytes,
                                         (_qSize(imageBits) - OHDR_SIZE));
#endif

            if (bytes > (_qSize(imageBits) - OHDR_SIZE)) {
                /* imageBits too small */
                XDestroyImage(image);
                goto fail;
            }
            if (image->bitmap_bit_order == MSBFirst)
                _ArrayInstPtr(info)->a_element[0] = @symbol(msbFirst);
            else
                _ArrayInstPtr(info)->a_element[0] = @symbol(lsbFirst);
            _ArrayInstPtr(info)->a_element[1] = _MKSMALLINT(image->depth);
            _ArrayInstPtr(info)->a_element[2] = _MKSMALLINT(image->bytes_per_line);
            bcopy(image->data, _ByteArrayInstPtr(imageBits)->ba_element, bytes);
            XDestroyImage(image);
            RETURN ( info );
        }
fail: ;
    }
%}
.
    self primitiveFailed.
    ^ nil
! !

!XWorkstation methodsFor:'drawing'!

displayString:aString x:x y:y in:aDrawableId with:aGCId round:round opaque:opaque 
    "draw a string - if opaque is false, draw foreground only; otherwise, draw both
     foreground and background characters.
     If the coordinates are not integers, the operation is retried if the argument round
     is true. If round is false, the x/y coordinates must be integers."

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Window win = _WindowVal(aDrawableId);
    unsigned char *cp;
    int n;
#ifdef TWOBYTESTRINGS
    extern OBJ TwoByteString;
#endif

    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(x) && _isSmallInteger(y)) {
        if (__isString(aString) || __isSymbol(aString)) {
            cp = _stringVal(aString);
            n = _stringSize(aString);
            if (n > 1000) n = 1000;
            if (opaque == true)
                XDrawImageString(myDpy, win, gc, _intVal(x), _intVal(y), (char *)cp, n);
            else
                XDrawString(myDpy, win, gc, _intVal(x), _intVal(y), (char *)cp, n);
            RETURN ( self );
        }
#ifdef TWOBYTESTRINGS
        if (_Class(aString) == TwoByteString) {
            cp = _stringVal(aString);
            n = _byteArraySize(aString) / 2;
            if (n > 1000) n = 1000;
            if (opaque == true)
                XDrawImageString16(myDpy, win, gc, _intVal(x), _intVal(y), (XChar2b *)cp, n);
            else
                XDrawString16(myDpy, win, gc, _intVal(x), _intVal(y), (XChar2b *)cp, n);
            RETURN ( self );
        }
#endif
    }
%}
.
    round ifTrue:[
        ^ self displayString:aString 
             x:x rounded 
             y:y rounded 
             in:aDrawableId 
             with:aGCId
             round:false
             opaque:opaque
    ].

    "x/y not integer, badGC or drawable, or not a string"
    self primitiveFailed
!

displayString:aString from:index1 to:index2 x:x y:y in:aDrawableId with:aGCId round:round opaque:opaque 
    "draw a sub-string - if opaque is false, draw foreground only; otherwise, draw both
     foreground and background characters.
     If the coordinates are not integers, the operation is retried if the argument round
     is true. If round is false, the x/y coordinates must be integers."

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Window win = _WindowVal(aDrawableId);
    unsigned char *cp;
    int  i1, i2, l, n;
#ifdef TWOBYTESTRINGS
    extern OBJ TwoByteString;
#endif

    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(index1) && _isSmallInteger(index2)
     && _isSmallInteger(x) && _isSmallInteger(y)) {
        if (__isString(aString) || __isSymbol(aString)) {
            i1 = _intVal(index1) - 1;
            i2 = _intVal(index2) - 1;
            n = _stringSize(aString);
            cp = _stringVal(aString);
            if ((i1 >= 0) && (i2 < n)) {
                if (i2 >= i1) {
                    cp += i1;
                    l = i2 - i1 + 1;
                    if (l > 1000) l = 1000;
                    if (opaque == true)
                        XDrawImageString(myDpy, win, gc, _intVal(x), _intVal(y), (char *)cp, l);
                    else
                        XDrawString(myDpy, win, gc, _intVal(x), _intVal(y), (char *)cp, l);
                }
                RETURN ( self );
            }
        }
#ifdef TWOBYTESTRINGS
        if (_Class(aString) == TwoByteString) {
            i1 = _intVal(index1) - 1;
            i2 = _intVal(index2) - 1;
            n = _byteArraySize(aString) / 2;
            cp = _stringVal(aString);
            if ((i1 >= 0) && (i2 < n)) {
                if (i2 >= i1) {
                    cp += (i1 * 2);
                    l = i2 - i1 + 1;
                    if (l > 1000) l = 1000;
                    if (opaque == true)
                        XDrawImageString16(myDpy, win, gc, _intVal(x), _intVal(y), (XChar2b *)cp, l);
                    else
                        XDrawString16(myDpy, win, gc, _intVal(x), _intVal(y), (XChar2b *)cp, l);
                    RETURN ( self );
                }
            }
        }
#endif
    }
%}
.
    round ifTrue:[
        ^ self displayString:aString 
             from:index1
             to:index2
             x:x rounded 
             y:y rounded 
             in:aDrawableId 
             with:aGCId
             round:false
             opaque:opaque
    ].

    "x/y not integer, badGC or drawable, or not a string"
    self primitiveFailed
!

displayPointX:x y:y in:aDrawableId with:aGCId
    "draw a point" 

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Window win = _WindowVal(aDrawableId);

    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(x) && _isSmallInteger(y)) {
        XDrawPoint(myDpy, win, gc, _intVal(x), _intVal(y));
        RETURN ( self );
    }
%}
.
    ^ self displayPointX:x rounded y:y rounded in:aDrawableId with:aGCId round:false
!

displayPointX:x y:y in:aDrawableId with:aGCId round:round
    "draw a point" 

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Window win = _WindowVal(aDrawableId);

    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(x) && _isSmallInteger(y)) {
        XDrawPoint(myDpy, win, gc, _intVal(x), _intVal(y));
        RETURN ( self );
    }
%}
.
    round ifTrue:[
        ^ self displayPointX:x rounded y:y rounded in:aDrawableId with:aGCId round:false
    ].

    "badGC, badDrawable or bad x/y"
    self primitiveFailed
!

displayLineFromX:x0 y:y0 toX:x1 y:y1 in:aDrawableId with:aGCId
    "draw a line" 

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Window win = _WindowVal(aDrawableId);

    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(x0) && _isSmallInteger(y0)
     && _isSmallInteger(x1) && _isSmallInteger(y1)) {
        XDrawLine(myDpy, win, gc, _intVal(x0), _intVal(y0),
                                  _intVal(x1), _intVal(y1));
        RETURN ( self );
    }
%}
.
    self displayRoundedLineFromX:x0 rounded y:y0 rounded toX:x1 rounded y:y1 rounded in:aDrawableId with:aGCId
!

displayRoundedLineFromX:x0 y:y0 toX:x1 y:y1 in:aDrawableId with:aGCId
    "draw a line; the points coordinates are already rounded to integers" 

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Window win = _WindowVal(aDrawableId);

    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(x0) && _isSmallInteger(y0)
     && _isSmallInteger(x1) && _isSmallInteger(y1)) {
        XDrawLine(myDpy, win, gc, _intVal(x0), _intVal(y0),
                                  _intVal(x1), _intVal(y1));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

displayRectangleX:x y:y width:width height:height in:aDrawableId with:aGCId
    "draw a rectangle" 

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Window win = _WindowVal(aDrawableId);
    int w, h;

    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(x) && _isSmallInteger(y)
     && _isSmallInteger(width) && _isSmallInteger(height)) {
        w = _intVal(width);
        h = _intVal(height);
        /*
         * need this check here: some servers simply dump core with bad args
         */
        if ((w >= 0) && (h >= 0)) {
            XDrawRectangle(myDpy, win, gc, _intVal(x), _intVal(y), w, h);
        }
        RETURN ( self );
    }
%}
.
    self displayRoundedRectangleX:x rounded y:y rounded width:width rounded height:height rounded in:aDrawableId with:aGCId
!

displayRoundedRectangleX:x y:y width:width height:height in:aDrawableId with:aGCId
    "draw a rectangle; the coordinates have already been rounded to integers" 

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Window win = _WindowVal(aDrawableId);
    int w, h;

    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(x) && _isSmallInteger(y)
     && _isSmallInteger(width) && _isSmallInteger(height)) {
        w = _intVal(width);
        h = _intVal(height);
        /*
         * need this check here: some servers simply dump core with bad args
         */
        if ((w >= 0) && (h >= 0)) {
            XDrawRectangle(myDpy, win, gc, _intVal(x), _intVal(y), w, h);
        }
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

displayPolygon:aPolygon in:aDrawableId with:aGCId
    "draw a polygon"

    |numberOfPoints index|

    numberOfPoints := aPolygon size.
%{
    {
        GC gc = _GCVal(aGCId);
        Window win = _WindowVal(aDrawableId);
        extern OBJ Point;
        OBJ point, x, y;
        int i, num;
        XPoint *points;
        XPoint qPoints[100];
        static struct inlineCache ix = _ILC0;
        static struct inlineCache iy = _ILC0;

        if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
         && _isSmallInteger(numberOfPoints)) {
            num = _intVal(numberOfPoints);
            if (num > 100)
                points = (XPoint *)malloc(sizeof(XPoint) * num);
            else
                points = qPoints;

            for (i=0; i<num; i++) {
                point = _AT_(aPolygon COMMA_CON, _MKSMALLINT(i+1));
                if (! __isPoint(point)) goto fail;
                x = _point_X(point);
                y = _point_Y(point);
                if (! _isSmallInteger(x) || ! _isSmallInteger(y))
                    goto fail;
                points[i].x = _intVal(x);
                points[i].y = _intVal(y);
            }
            XDrawLines(myDpy, win, gc, points, num, CoordModeOrigin);
            if (num > 100)
                free(points);
            RETURN ( self );
        }
fail: ;
    }
%}
.
    self primitiveFailed
!

copyFromFaxImage:sourceId x:srcX y:srcY to:destId x:dstX y:dstY 
                      width:w height:h with:aGCId scaleX:scaleX scaleY:scaleY

    "do a bit-blt of a compressed FAX image 
     - this needs my private FAX extension"
%{
#ifdef FAX
    GC gc = _GCVal(aGCId);
    FAXImage source;
    Drawable dest;

    if (_isSmallInteger(aGCId) && _isSmallInteger(sourceId)
     && _isSmallInteger(destId)
     && _isSmallInteger(scaleX) && _isSmallInteger(scaleY)
     && _isSmallInteger(w) && _isSmallInteger(h)
     && _isSmallInteger(srcX) && _isSmallInteger(srcY)
     && _isSmallInteger(dstX) && _isSmallInteger(dstY)) {
        source = (FAXImage) _WindowVal(sourceId);
        dest = (Drawable) _WindowVal(destId);
        XFAXImageCopyImage(myDpy, source, dest, gc,
                                _intVal(srcX), _intVal(srcY),
                                _intVal(w), _intVal(h),
                                _intVal(dstX), _intVal(dstY),
                                _intVal(scaleX), _intVal(scaleY));
        RETURN ( self );
    }
#endif
%}
.
    self primitiveFailed
!

copyFromId:sourceId x:srcX y:srcY to:destId x:dstX y:dstY 
                width:w height:h with:aGCId
    "do a bit-blt"

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Drawable source, dest;

    if (_isSmallInteger(aGCId) && _isSmallInteger(sourceId)
     && _isSmallInteger(destId)
     && _isSmallInteger(w) && _isSmallInteger(h)
     && _isSmallInteger(srcX) && _isSmallInteger(srcY)
     && _isSmallInteger(dstX) && _isSmallInteger(dstY)) {
        source = (Drawable) _WindowVal(sourceId);
        dest = (Drawable) _WindowVal(destId);
        XCopyArea(myDpy, source, dest, gc,
                                 _intVal(srcX), _intVal(srcY),
                                _intVal(w), _intVal(h),
                                _intVal(dstX), _intVal(dstY));
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

copyPlaneFromId:sourceId x:srcX y:srcY to:destId x:dstX y:dstY 
                width:w height:h with:aGCId
    "do a bit-blt"

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Drawable source, dest;

    if (_isSmallInteger(aGCId) && _isSmallInteger(sourceId)
     && _isSmallInteger(destId)
     && _isSmallInteger(w) && _isSmallInteger(h)
     && _isSmallInteger(srcX) && _isSmallInteger(srcY)
     && _isSmallInteger(dstX) && _isSmallInteger(dstY)) {
        source = (Drawable) _WindowVal(sourceId);
        dest = (Drawable) _WindowVal(destId);
        XCopyPlane(myDpy, source, dest, gc,
                                 _intVal(srcX), _intVal(srcY),
                                 _intVal(w), _intVal(h),
                                 _intVal(dstX), _intVal(dstY), 1);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

displayArcX:x y:y w:width h:height from:startAngle angle:angle
             in:aDrawableId with:aGCId
    "draw an arc"

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Window win = _WindowVal(aDrawableId);
    int w, h, angle1, angle2;
    double f;

    if (_isSmallInteger(startAngle))
        angle1 = _intVal(startAngle) * 64;
    else if (__isFloat(startAngle)) {
        f = _floatVal(startAngle);
        angle1 = f * 64;
    }
    if (_isSmallInteger(angle))
        angle2 = _intVal(angle) * 64;
    else if (__isFloat(angle)) {
        f = _floatVal(angle);
        angle2 = f * 64;
    }
    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(x) && _isSmallInteger(y)
     && _isSmallInteger(width) && _isSmallInteger(height)) {
        w = _intVal(width);
        h = _intVal(height);
        /*
         * need this check here: some servers simply dump core with bad args
         */
        if ((w >= 0) && (h >= 0) && (angle1 >= 0) && (angle2 >= 0)) {
            XDrawArc(myDpy, win, gc, _intVal(x), _intVal(y),
                                   w, h, angle1, angle2);
        }
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

fillArcX:x y:y w:width h:height from:startAngle angle:angle
               in:aDrawableId with:aGCId
    "fill an arc"

%{  /* NOCONTEXT */

    GC gc = _GCVal(aGCId);
    Window win = _WindowVal(aDrawableId);
    int w, h, angle1, angle2;
    double f;

    if (_isSmallInteger(startAngle))
        angle1 = _intVal(startAngle) * 64;
    else if (__isFloat(startAngle)) {
        f = _floatVal(startAngle);
        angle1 = f * 64;
    }
    if (_isSmallInteger(angle))
        angle2 = _intVal(angle) * 64;
    else if (__isFloat(angle)) {
        f = _floatVal(angle);
        angle2 = f * 64;
    }
    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(x) && _isSmallInteger(y)
     && _isSmallInteger(width) && _isSmallInteger(height)) {
        w = _intVal(width);
        h = _intVal(height);
        /*
         * need this check here: some servers simply dump core with bad args
         */
        if ((w >= 0) && (h >= 0) && (angle1 >= 0) && (angle2 >= 0)) {
            XFillArc(myDpy, win, gc, _intVal(x), _intVal(y),
                                   w, h, angle1, angle2);
        }
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

fillRectangleX:x y:y width:width height:height in:aDrawableId with:aGCId
    "fill a rectangle"

%{  /* NOCONTEXT */

    int w, h;

    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(x) && _isSmallInteger(y)
     && _isSmallInteger(width) && _isSmallInteger(height)) {
        w = _intVal(width);
        h = _intVal(height);
        /*
         * need this check here: some servers simply dump core with bad args
         */
        if ((w >= 0) && (h >= 0)) {
            XFillRectangle(myDpy,
                           (Drawable)_WindowVal(aDrawableId), _GCVal(aGCId),
                           _intVal(x), _intVal(y), w, h);
        }
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

fillPolygon:aPolygon in:aDrawableId with:aGCId
    "fill a polygon given by its points"

    |numberOfPoints index|

    numberOfPoints := aPolygon size.
%{
    {
        GC gc = _GCVal(aGCId);
        Window win = _WindowVal(aDrawableId);
        OBJ point, x, y;
        extern OBJ Point;
        int i, num;
        XPoint *points;
        XPoint qPoints[100];
        static struct inlineCache ix = _ILC0;
        static struct inlineCache iy = _ILC0;

        if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
         && _isSmallInteger(numberOfPoints)) {
            num = _intVal(numberOfPoints);
            if (num < 3) {
                RETURN ( self );
            }
            if (num > 100)
                points = (XPoint *) malloc(sizeof(XPoint) * num);
            else
                points = qPoints;
            for (i=0; i<num; i++) {
                point = _AT_(aPolygon COMMA_CON, _MKSMALLINT(i+1));
                if (! __isPoint(point)) goto fail;
                x = _point_X(point);
                y = _point_Y(point);
                if (! _isSmallInteger(x) || ! _isSmallInteger(y))
                    goto fail;
                points[i].x = _intVal(x);
                points[i].y = _intVal(y);
            }
            XFillPolygon(myDpy, win, gc, points, num, Complex, CoordModeOrigin);
            if (num > 100)
                free(points);
            RETURN ( self );
        }
fail: ;
    }
%}
.
    self primitiveFailed
!

drawBits:imageBits depth:imageDepth 
                   width:imageWidth height:imageHeight 
                       x:srcx y:srcy
                    into:aDrawableId 
                       x:dstx y:dsty 
                   width:w height:h 
                    with:aGCId

    ^ self drawBits:imageBits bitsPerPixel:imageDepth depth:imageDepth 
                                     width:imageWidth height:imageHeight 
                                         x:srcx y:srcy
                                      into:aDrawableId 
                                         x:dstx y:dsty 
                                     width:w height:h 
                                      with:aGCId

    "draw a bitimage which has depth id, width iw and height ih into
     the drawable. draw a region of w/h pixels from srcx/srcy to dstx/dsty.
     Individual source pixels must have imageDepth bits.
     It has to be checked elsewhere, that the server can do it with the given
     depth - otherwise, primitive failure will be signalled.
     Also it is assumed, that the colormap is setup correctly and the
     colors are allocated - otherwise the colors may be wrong."

!

drawBits:imageBits bitsPerPixel:bitsPerPixel depth:imageDepth 
                          width:imageWidth height:imageHeight 
                              x:srcx y:srcy
                           into:aDrawableId 
                              x:dstx y:dsty 
                          width:w height:h 
                           with:aGCId

    "draw a bitimage which has depth id, width iw and height ih into
     the drawable. draw a region of w/h pixels from srcx/srcy to dstx/dsty.
     Individual source pixels have bitsPerPixel bits, allowing to draw
     depth and pixel-units to be different.
     It has to be checked elsewhere, that the server can do it with the given
     depth - otherwise, primitive failure will be signalled.
     Also it is assumed, that the colormap is setup correctly and the
     colors are allocated - otherwise the colors may be wrong."

    "
     sorry; I had to separate it into 2 methods, since XPutImage needs
     an unlimited stack, and thus cannot send primitiveFailed
    "
    (self primDrawBits:imageBits bitsPerPixel:bitsPerPixel depth:imageDepth 
                                        width:imageWidth height:imageHeight 
                                             x:srcx y:srcy
                                          into:aDrawableId 
				             x:dstx y:dsty 
				         width:w height:h 
                                          with:aGCId)
    ifFalse:[
	"
	 also happens, if a segmentation violation occurs in the 
	 XPutImage ...
	"
        self primitiveFailed
    ].
!

primDrawBits:imageBits bitsPerPixel:bitsPerPixel depth:imageDepth 
	                      width:imageWidth height:imageHeight 
                                  x:srcx y:srcy
                               into:aDrawableId 
                                  x:dstx y:dsty 
			      width:w height:h 
			       with:aGCId

%{  /* UNLIMITEDSTACK */

    /*
     * need unlimited stack, since some Xlibs do a huge alloca in
     * XPutImage
     */
    Display *dpy = myDpy;
    GC gc = _GCVal(aGCId);
    Window win = _WindowVal(aDrawableId);
    extern OBJ ByteArray;
    XImage image;

    if (_isSmallInteger(aGCId) && _isSmallInteger(aDrawableId)
     && _isSmallInteger(srcx) && _isSmallInteger(srcy)
     && _isSmallInteger(dstx) && _isSmallInteger(dsty)
     && _isSmallInteger(w) && _isSmallInteger(h)
     && _isSmallInteger(imageWidth) && _isSmallInteger(imageHeight)
     && _isSmallInteger(imageDepth) && _isSmallInteger(bitsPerPixel)
     && __isByteArray(imageBits)) {
        image.data = (char *)_ByteArrayInstPtr(imageBits)->ba_element;
        image.width = _intVal(imageWidth);
        image.height = _intVal(imageHeight);
        image.xoffset = 0;
        image.format = ZPixmap;
        image.byte_order = MSBFirst;
        image.bitmap_unit = 8;
        image.bitmap_bit_order = MSBFirst;
        image.bitmap_pad = 8;
        image.depth = _intVal(imageDepth);
        image.bits_per_pixel = _intVal(bitsPerPixel);
        switch (image.bits_per_pixel) {
            case 1:
                image.bytes_per_line = (_intVal(imageWidth) + 7) / 8;
                break;
            case 2:
                image.bytes_per_line = (_intVal(imageWidth)*2 + 7) / 8;
                break;
            case 4:
                image.bytes_per_line = (_intVal(imageWidth)*4 + 7) / 8;
                break;
            case 8:
                image.bytes_per_line = _intVal(imageWidth);
                break;
            case 16:
                image.bytes_per_line = _intVal(imageWidth)*2;
                break;
            case 24:
                image.bytes_per_line = _intVal(imageWidth)*3;
                break;
            default:
                goto fail;
        }
        image.red_mask = 0xFFFF;
        image.green_mask = 0xFFFF;
        image.blue_mask = 0xFFFF;
        XPutImage(dpy, win, gc, &image, _intVal(srcx), _intVal(srcy),
                                        _intVal(dstx), _intVal(dsty),
                                        _intVal(w), _intVal(h));
        RETURN ( true );
    }
fail: ;
%}
.
    ^ false
! !

!XWorkstation methodsFor:'events'!

eventMaskFor:anEventSymbol
    "return the eventMask bit-constant corresponding to an event symbol"

    (anEventSymbol == #keyPress) ifTrue:[^ KeyPressMask].
    (anEventSymbol == #keyRelease) ifTrue:[^ KeyPressMask].
    (anEventSymbol == #buttonPress) ifTrue:[^ ButtonPressMask].
    (anEventSymbol == #buttonRelease) ifTrue:[^ ButtonReleaseMask].
    (anEventSymbol == #buttonMotion) ifTrue:[^ ButtonMotionMask].
    (anEventSymbol == #pointerMotion) ifTrue:[^ PointerMotionMask].
    (anEventSymbol == #expose) ifTrue:[^ ExposureMask].
    (anEventSymbol == #focusChange) ifTrue:[^ FocusChangeMask].
    (anEventSymbol == #enter) ifTrue:[^ EnterWindowMask].
    (anEventSymbol == #leave) ifTrue:[^ LeaveWindowMask].
    ^ 0
!

setEventMask:aMask in:aWindowId
    "tell X that we are only interrested in events from aMask, which
     is the bitwise or of the eventMask bits (see 'eventMaskFor:')"

%{  /* NOCONTEXT */

    int mask;

    if (_isSmallInteger(aWindowId) && _isSmallInteger(aMask)) {
        mask = _intVal(aMask);

        /* these may not be disabled */
        mask |= ExposureMask | StructureNotifyMask /*| SubstructureNotifyMask */;

        XSelectInput(myDpy, _WindowVal(aWindowId), mask);
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

startDispatch
    "redefined to clear dispatchingExpose, which is a special X feature"

    dispatching ifTrue:[^ self].
    dispatchingExpose := nil.
    super startDispatch
!

checkForEndOfDispatch
    "return true, if there are still any views of interrest - if not,
     stop dispatch"

"/    idToViewMapping isEmpty ifTrue:[
"/	dispatching := false.
"/    ] ifFalse:[
"/       ((idToViewMapping size == 1) and:[idToViewMapping includesValue:RootView]) ifTrue:[
"/	   dispatching := false.
"/	]
"/    ]

    knownViews isEmpty ifTrue:[
        dispatching := false.
        ^ self
    ].
    ((knownViews size == 1) and:[(knownViews at:1) == RootView]) ifTrue:[
        dispatching := false.
    ]
!

dispatchPendingEvents
    "central event handling method:
     this code is somewhat special, since X has a concept of graphic
     expose events (which are sent after a bitblt). After such a bitblt,
     we only handle exposes until the graphicsExpose arrives.
     Other systems may not need such a kludge"

    "interrested in exposes only ?"
    dispatchingExpose notNil ifTrue:[
        [self exposeEventPendingFor:dispatchingExpose] whileTrue:[
            self dispatchExposeEventFor:dispatchingExpose
        ].
        ^ self
    ].

    "no general dispatch"
    [self eventPendingWithoutSync] whileTrue:[
        self dispatchEventFor:nil withMask:nil
    ]
!

handleExposeOnlyFor:aView
    "from now on, handle expose events only"

    dispatchingExpose := aView id
!

handleAllEvents
    "from now on, handle any kind of event"

    dispatchingExpose := nil
!

dispatchExposeEventFor:aViewIdOrNil
    "get next expose event and send appropriate message to the view or the controller
     (if the view has one).
     If the argument aViewIdOrNil is nil, events for any view are processed,
     otherwise only events for the view with given id are processed."
     
    self dispatchEventFor:aViewIdOrNil withMask:(self eventMaskFor:#expose)
!

dispatchEventFor:aViewIdOrNil withMask:eventMask
    "central event handling method:
     get next event and send appropriate message to the view or the controller
     (if the view has one).
     If the argument aViewIdOrNil is nil, events for any view are processed,
     otherwise only events for the view with given id are processed.
     If the argument aMask is nonNil, only events for this eventMask are
     handled."

    self getEventFor:aViewIdOrNil withMask:eventMask.
    self dispatchLastEvent.
    Processor yield
!

getEventFor:aViewIdOrNil withMask:eventMask
    "read next event - put into local buffer - sorry had to split
     dispatch into this get and a handle to allow unlimitedstack here
     (some Xlibs do a big alloca there ...)"

%{  /* UNLIMITEDSTACK */

    Display *dpy = myDpy;
    Window win, wWanted;
    extern int _immediateInterrupt;

    ev.type = 0;
    /* _immediateInterrupt = 1; */
    if (_isSmallInteger(aViewIdOrNil)) {
        wWanted = _WindowVal(aViewIdOrNil);
        if (_isSmallInteger(eventMask)) 
            XWindowEvent(dpy, wWanted, _intVal(eventMask), &ev);
        else
            XWindowEvent(dpy, wWanted, ~0, &ev);
    } else {
        if (_isSmallInteger(eventMask)) 
            XMaskEvent(dpy, _intVal(eventMask), &ev);
        else
            XNextEvent(dpy, &ev);
    }
    /* _immediateInterrupt = 0; */
%}
!

dispatchLastEvent
    |theView controller sensor delegate keyButtonReceiver exposeReceiver
     sym symS arg butt|

%{  /* STACK: 8000 */
#   define ANYBUTTON   (Button1MotionMask | Button2MotionMask | Button3MotionMask)

    Display *dpy = myDpy;
    Window win, wWanted;
    XAnyEvent *ae = (XAnyEvent *)&ev;
#   define ee ((XExposeEvent *)&ev)
#   define ke ((XKeyPressedEvent *)&ev)
#   define be ((XButtonPressedEvent *)&ev)
#   define ce ((XConfigureEvent *)&ev)
#   define me ((XMotionEvent *)&ev)
#   define ewe ((XEnterWindowEvent *)&ev)
#   define lwe ((XLeaveWindowEvent *)&ev)
#   define de ((XDestroyWindowEvent *)&ev)
    KeySym keySym;
    unsigned char buffer[10];
    int i, nchars;
    char *keySymString;
    char keySymStringBuffer[32];
    static unsigned multiClickTime = 0;
    unsigned nextMultiClickTime;
    struct inlineCache *ip, *ipS;
    static struct inlineCache vid = _ILC1;
    static struct inlineCache exp = _ILC4;
    static struct inlineCache gexp = _ILC4;
    static struct inlineCache nexp = _ILC0;
    static struct inlineCache conf = _ILC4;
    static struct inlineCache unmap = _ILC0;
    static struct inlineCache map = _ILC0;
    static struct inlineCache mot = _ILC3;
    static struct inlineCache kp = _ILC3;
    static struct inlineCache bp = _ILC3;
    static struct inlineCache bmp = _ILC3;
    static struct inlineCache bsp = _ILC3;
    static struct inlineCache br = _ILC3;
    static struct inlineCache focIn = _ILC0;
    static struct inlineCache focOut = _ILC0;
    static struct inlineCache pe = _ILC3;
    static struct inlineCache pl = _ILC1;
    static struct inlineCache term = _ILC0;
    static struct inlineCache savterm = _ILC0;
    static struct inlineCache destr = _ILC0;

    static struct inlineCache expS = _ILC5;
    static struct inlineCache gexpS = _ILC5;
    static struct inlineCache nexpS = _ILC1;
    static struct inlineCache motS = _ILC4;
    static struct inlineCache kpS = _ILC4;
    static struct inlineCache bpS = _ILC4;
    static struct inlineCache bmpS = _ILC4;
    static struct inlineCache bspS = _ILC4;
    static struct inlineCache brS = _ILC4;
    static struct inlineCache focInS = _ILC1;
    static struct inlineCache focOutS = _ILC1;
    static struct inlineCache peS = _ILC4;
    static struct inlineCache plS = _ILC2;

    static struct inlineCache keymap = _ILC0;
    static struct inlineCache created = _ILC0;
    static struct inlineCache mapReq = _ILC0;
    static struct inlineCache repar = _ILC0;
    static struct inlineCache confReq = _ILC0;
    static struct inlineCache resReq = _ILC0;
    static struct inlineCache prop = _ILC0;
    static struct inlineCache selClear = _ILC0;
    static struct inlineCache selReq = _ILC0;
    static struct inlineCache selNotify = _ILC0;
    static struct inlineCache colormap = _ILC0;
    static struct inlineCache vis = _ILC0;

    static struct inlineCache contr = _ILC0;
    static struct inlineCache sens = _ILC0;
    static struct inlineCache deleg = _ILC0;
    static struct inlineCache skp = _ILC4;
    static struct inlineCache skpS = _ILC4;

    theView = (*vid.ilc_func)(self, @symbol(viewFromId:) COMMA_CON, nil, &vid, 
                                    MKOBJ(ae->window));

    if (theView == nil) {
        RETURN (nil);
    }

    /*
     * key, button and pointer events are sent to the sensor,
     * or (if there is none) to the delegate.
     * or (if there is none) to the view.
     *
     * expose events are sent to the sensor,
     * or (if there is none) to the view
     *
     * sensor and delegate get view as additional argument
     */
    keyButtonReceiver = nil;
    exposeReceiver = nil;
    sensor = (*sens.ilc_func)(theView, @symbol(sensor) COMMA_CON, nil, &sens);
    if (sensor != nil) {
        keyButtonReceiver = sensor;
        exposeReceiver = sensor;
    } else {
        delegate = (*deleg.ilc_func)(theView, @symbol(delegate) COMMA_CON, nil, &deleg);
        if (delegate != nil) {
            keyButtonReceiver = delegate;
        }
    }

    /*
     * now, receiver is in variables keyButtonReceiver, exposeReceiver (to send with view)
     * or theView if not.
     */
    switch (ev.type) {
            /*
             * I do not support key-release
             * to press everyone NOT to use stuff which might not be
             * available on all machines ...
             */
        case KeyRelease:
            _INST(altDown) = (ke->state & Mod2Mask) ? true : false;
            _INST(metaDown) = (ke->state & Mod1Mask) ? true : false;
            _INST(shiftDown) = (ke->state & ShiftMask) ? true : false;
            _INST(controlDown) = (ke->state & ControlMask) ? true : false;
            break;

        case KeyPress:
            _INST(eventRootX) = _MKSMALLINT(ke->x_root);
            _INST(eventRootY) = _MKSMALLINT(ke->y_root);
            _INST(altDown) = (ke->state & Mod2Mask) ? true : false;
            _INST(metaDown) = (ke->state & Mod1Mask) ? true : false;
            _INST(shiftDown) = (ke->state & ShiftMask) ? true : false;
            _INST(controlDown) = (ke->state & ControlMask) ? true : false;

            arg = nil;
            nchars = XLookupString(ke, (char *)buffer, sizeof(buffer), &keySym, NULL);
            if (nchars 
             && (((buffer[0] >= ' ') && (buffer[0] <= '~'))
                 || (buffer[0] >= 0x80))) {
                arg = *_CharacterTable[buffer[0]];
                keySymString = NULL;
            } else {
                keySymString = XKeysymToString(keySym);
                if (keySymString) {
                    if (keySymString[0] == 'D') {
                        /*
                         * remove underscore, dont want it in symbols
                         */
                        if (strcmp(keySymString, "Delete_line") == 0) {
                            keySymString = "DeleteLine";
                        } else if (strcmp(keySymString, "Delete_word") == 0) {
                            keySymString = "DeleteWord";
                        }
                    }
                    /*
                     * make names compatible
                     */
                    if (strcmp(keySymString, "Down") == 0) {
                        keySymString = "CursorDown";
                    } else if (strcmp(keySymString, "Up") == 0) {
                        keySymString = "CursorUp";
                    } else if (strcmp(keySymString, "Left") == 0) {
                        keySymString = "CursorLeft";
                    } else if (strcmp(keySymString, "Right") == 0) {
                        keySymString = "CursorRight";
                    }
                    arg = _MKSYMBOL(keySymString, (OBJ *)0, __context);
                }
            }

            if (arg == nil) {
                /* happens sometimes (alt-graph on sun has no keysym) */
                break;
            }

            if (keyButtonReceiver != nil) {
                (*skpS.ilc_func)(keyButtonReceiver, 
                                 @symbol(keyPress:x:y:view:)
                                 COMMA_CON, nil, &skpS,
                                 arg, 
                                 _MKSMALLINT(ke->x), 
                                 _MKSMALLINT(ke->y),
                                 theView);
                break;
            }

            (*skp.ilc_func)(self, 
                            @symbol(sendKeyPress:x:y:to:)
                            COMMA_CON, nil, &skp,
                            arg, 
                            _MKSMALLINT(ke->x), 
                            _MKSMALLINT(ke->y),
                            theView);
            break;

        case ButtonPress:
            _INST(buttonsPressed) = _MKSMALLINT(_intVal(_INST(buttonsPressed)) | (1 << be->button));
            _INST(eventRootX) = _MKSMALLINT(be->x_root);
            _INST(eventRootY) = _MKSMALLINT(be->y_root);

            if (_isSmallInteger(DeviceWorkstation_MultiClickTimeDelta))
                nextMultiClickTime = be->time + _intVal(DeviceWorkstation_MultiClickTimeDelta);
            else
                nextMultiClickTime = 0;

            if (multiClickTime) {
                if (be->time < multiClickTime) {
                    multiClickTime = nextMultiClickTime;
                    ip = &bmp;
                    ipS = &bmpS;
                    sym = @symbol(buttonMultiPress:x:y:);
                    symS = @symbol(buttonMultiPress:x:y:view:);
                    goto sendButtonEvent;
                    break;
                }
            }
            multiClickTime = nextMultiClickTime;
            if (be->state & ShiftMask) {
                ip = &bsp;
                ipS = &bspS;
                sym = @symbol(buttonShiftPress:x:y:);
                symS = @symbol(buttonShiftPress:x:y:view:);
                goto sendButtonEvent;
            }
            ip = &bp;
            ipS = &bpS;
            sym = @symbol(buttonPress:x:y:);
            symS = @symbol(buttonPress:x:y:view:);
            goto sendButtonEvent;
            
            /* NOT REACHED */

        case ButtonRelease:
            _INST(buttonsPressed) = _MKSMALLINT(_intVal(_INST(buttonsPressed)) & ~(1 << be->button));
            _INST(eventRootX) = _MKSMALLINT(be->x_root);
            _INST(eventRootY) = _MKSMALLINT(be->y_root);
            ip = &br;
            ipS = &brS;
            sym = @symbol(buttonRelease:x:y:);
            symS = @symbol(buttonRelease:x:y:view:);
            /* fall into */

        sendButtonEvent:
            butt = _MKSMALLINT(be->button);
#ifdef NOTDEF
            /*
             * this allows operation with single button mouses: meta-click is always Button 2
             */
            if (_INST(metaDown) == true)
                butt = _MKSMALLINT(2);
            else 
#endif
                butt = _AT_(DeviceWorkstation_ButtonTranslation, butt);
            

            if (keyButtonReceiver != nil) {
                (*(*ipS).ilc_func)(keyButtonReceiver, 
                                   symS
                                   COMMA_CON, nil, ipS,
                                   butt, 
                                   _MKSMALLINT(ke->x), 
                                   _MKSMALLINT(ke->y),
                                   theView);
                break;
            }

            (*(*ip).ilc_func)(theView, 
                              sym 
                              COMMA_CON, nil, ip,
                              butt,
                              _MKSMALLINT(be->x), 
                              _MKSMALLINT(be->y));
            break;

        case MotionNotify:
            if (_INST(motionEventCompression) != false) {
                while (XCheckWindowEvent(dpy, me->window, ANYBUTTON, &ev)) ;;
            }

            _INST(eventRootX) = _MKSMALLINT(me->x_root);
            _INST(eventRootY) = _MKSMALLINT(me->y_root);

            if (keyButtonReceiver != nil) {
                (*motS.ilc_func)(keyButtonReceiver, 
                                 @symbol(buttonMotion:x:y:view:)
                                 COMMA_CON, nil, &motS,
                                 _MKSMALLINT(me->state),
                                 _MKSMALLINT(me->x),
                                 _MKSMALLINT(me->y),
                                 theView);
                break;
            }
            (*mot.ilc_func)(theView, 
                            @symbol(buttonMotion:x:y:)
                            COMMA_CON, nil, &mot,
                            _MKSMALLINT(me->state),
                            _MKSMALLINT(me->x),
                            _MKSMALLINT(me->y));
            break;

        case FocusIn:
            if (keyButtonReceiver != nil) {
                (*focInS.ilc_func)(keyButtonReceiver, 
                                   @symbol(focusInView:)
                                   COMMA_CON, nil, &focInS, 
                                   theView);
                break;
            }
            (*focIn.ilc_func)(theView, 
                              @symbol(focusIn)
                              COMMA_CON, nil, &focIn);
            break;

        case FocusOut:
            if (keyButtonReceiver != nil) {
                (*focOutS.ilc_func)(keyButtonReceiver, 
                                    @symbol(focusOutView:)
                                    COMMA_CON, nil, &focOutS, 
                                    theView);
                break;
            }
            (*focOut.ilc_func)(theView, 
                               @symbol(focusOut)
                               COMMA_CON, nil, &focOut);
            break;

        case EnterNotify:
            if (keyButtonReceiver != nil) {
                (*peS.ilc_func)(keyButtonReceiver, 
                                @symbol(pointerEnter:x:y:view:)
                                COMMA_CON, nil, &peS,
                                _MKSMALLINT(ewe->state),
                                _MKSMALLINT(ewe->x), 
                                _MKSMALLINT(ewe->y),
                                theView);
                break;
            }
            (*pe.ilc_func)(theView, 
                           @symbol(pointerEnter:x:y:)
                           COMMA_CON, nil, &pe,
                           _MKSMALLINT(ewe->state),
                           _MKSMALLINT(ewe->x), 
                           _MKSMALLINT(ewe->y));
            break;

        case LeaveNotify:
            if (keyButtonReceiver != nil) {
                (*plS.ilc_func)(keyButtonReceiver, 
                                @symbol(pointerLeave:view:)
                                COMMA_CON, nil, &plS,
                                _MKSMALLINT(lwe->state), 
                                theView);
                break;
            }
            (*pl.ilc_func)(theView, 
                           @symbol(pointerLeave:)
                           COMMA_CON, nil, &pl,
                           _MKSMALLINT(lwe->state));
            break;

        case GraphicsExpose:
            if (exposeReceiver != nil) {
                (*gexpS.ilc_func)(exposeReceiver, 
                                  @symbol(graphicExposeX:y:width:height:view:)
                                  COMMA_CON, nil, &gexpS,
                                  _MKSMALLINT(ee->x),
                                  _MKSMALLINT(ee->y),
                                  _MKSMALLINT(ee->width),
                                  _MKSMALLINT(ee->height),
                                  theView);
                if (ee->count == 0) {
                    (*nexpS.ilc_func)(exposeReceiver, 
                                      @symbol(noExposeView:)
                                      COMMA_CON, nil, &nexpS, 
                                      theView);
                }
                break;
            }
            (*gexp.ilc_func)(theView, 
                             @symbol(graphicExposeX:y:width:height:)
                             COMMA_CON, nil, &gexp,
                             _MKSMALLINT(ee->x),
                             _MKSMALLINT(ee->y),
                             _MKSMALLINT(ee->width),
                             _MKSMALLINT(ee->height));
            if (ee->count == 0) {
                (*nexp.ilc_func)(theView, 
                                 @symbol(noExpose)
                                 COMMA_CON, nil, &nexp);
            }
            break;

        case Expose:
            if (exposeReceiver != nil) {
                (*expS.ilc_func)(exposeReceiver, 
                                 @symbol(exposeX:y:width:height:view:)
                                 COMMA_CON, nil, &expS,
                                 _MKSMALLINT(ee->x),
                                 _MKSMALLINT(ee->y),
                                 _MKSMALLINT(ee->width),
                                 _MKSMALLINT(ee->height),
                                 theView);
                break;
            }
            (*exp.ilc_func)(theView, 
                            @symbol(exposeX:y:width:height:)
                            COMMA_CON, nil, &exp,
                            _MKSMALLINT(ee->x),
                            _MKSMALLINT(ee->y),
                            _MKSMALLINT(ee->width),
                            _MKSMALLINT(ee->height));
            break;

        case NoExpose:
            if (exposeReceiver != nil) {
                (*nexpS.ilc_func)(exposeReceiver, 
                                  @symbol(noExposeView:)
                                  COMMA_CON, nil, &nexpS, 
                                  theView);
                break;
            }
            (*nexp.ilc_func)(theView, 
                             @symbol(noExpose) 
                             COMMA_CON, nil, &nexp);
            break;

        case ConfigureNotify:
            (*conf.ilc_func)(theView, 
                             @symbol(configureX:y:width:height:)
                             COMMA_CON, nil, &conf,
                             _MKSMALLINT(ce->x),
                             _MKSMALLINT(ce->y),
                             _MKSMALLINT(ce->width),
                             _MKSMALLINT(ce->height));
            break;

        case ClientMessage:
            if (ev.xclient.message_type == (int) _AtomVal(_INST(ProtocolsAtom))) {
                if (ev.xclient.data.l[0] == (int) _AtomVal(_INST(QuitAppAtom))) {
                    (*term.ilc_func)(theView, 
                                     @symbol(terminate)
                                     COMMA_CON, nil, &term);
                    break;
                }
                if (ev.xclient.data.l[0] == (int) _AtomVal(_INST(DeleteWindowAtom))) {
                    (*term.ilc_func)(theView, 
                                     @symbol(terminate)
                                     COMMA_CON, nil, &term);
                    break;
                }
                if (ev.xclient.data.l[0] == (int) _AtomVal(_INST(SaveYourselfAtom))) {
                    (*savterm.ilc_func)(theView, 
                                        @symbol(saveAndTerminate)
                                        COMMA_CON, nil, &savterm);
                    break;
                }
            }
            break;

        case DestroyNotify:
            (*destr.ilc_func)(theView, @symbol(destroyed) COMMA_CON, nil, &destr);
            break;

        case UnmapNotify:
            (*unmap.ilc_func)(theView, @symbol(unmapped) COMMA_CON, nil, &unmap);
            break;

        case MapNotify:
            (*map.ilc_func)(theView, @symbol(mapped) COMMA_CON, nil, &map);
            break;

        case KeymapNotify:
            (*keymap.ilc_func)(theView, @symbol(keyMapChange) COMMA_CON, nil, &keymap);
            break;

        case VisibilityNotify:
            (*vis.ilc_func)(theView, @symbol(visibilityChange) COMMA_CON, nil, &vis);
            break;

        case CreateNotify:
            (*created.ilc_func)(theView, @symbol(created) COMMA_CON, nil, &created);
            break;

        case MapRequest:
            (*mapReq.ilc_func)(theView, @symbol(mapRequest) COMMA_CON, nil, &mapReq);
            break;

        case ReparentNotify:
            (*repar.ilc_func)(theView, @symbol(reparented) COMMA_CON, nil, &repar);
            break;

        case ConfigureRequest:
            (*confReq.ilc_func)(theView, @symbol(configRequest) COMMA_CON, nil, &confReq);
            break;

        case GravityNotify:
        case ResizeRequest:
            (*resReq.ilc_func)(theView, @symbol(resizeRequest) COMMA_CON, nil, &resReq);
            break;

        case CirculateNotify:
        case CirculateRequest:
            break;

        case PropertyNotify:
            (*prop.ilc_func)(theView, @symbol(propertyChange) COMMA_CON, nil, &prop);
            break;

        case SelectionClear:
            (*selClear.ilc_func)(theView, @symbol(selectionClear) COMMA_CON, nil, &selClear);
            break;

        case SelectionRequest:
            (*selReq.ilc_func)(theView, @symbol(selectionRequest) COMMA_CON, nil, &selReq);
            break;

        case SelectionNotify:
            (*selNotify.ilc_func)(theView, @symbol(selectionChange) COMMA_CON, nil, &selNotify);
            break;

        case ColormapNotify:
            (*colormap.ilc_func)(theView, @symbol(colorMapChange) COMMA_CON, nil, &colormap);
            break;

        case MappingNotify:
printf("mapping\n");
/*
            (*mapping.ilc_func)(theView, @symbol(mapping) COMMA_CON, nil, &mapping);
*/
            break;
    }
#undef ee
#undef ke
#undef be
#undef ce
#undef me
#undef ewe
#undef lwe
#undef de
%}
.
    ^ true
!

disposeEventsWithMask:aMask for:aWindowId
    "dispose (throw away) specific events"

%{  /* NOCONTEXT */

    Display *dpy = myDpy;
    XEvent ev;
    Window win;

    if (_isSmallInteger(aMask)) {
        win = _WindowVal(aWindowId);
        while (XCheckWindowEvent(dpy, win, _intVal(aMask), &ev)) ;;
        RETURN ( self );
    }
%}
.
    self primitiveFailed
!

eventPending
    "return true, if any event is pending"

    dispatchingExpose notNil ifTrue:[
        ^ self exposeEventPendingFor:dispatchingExpose
    ].
    ^ self eventPendingWithSync
!

eventPendingWithSync
    "return true, if any event is pending. 
     Do a flush output buffer before."

%{  /* UNLIMITEDSTACK */

    Display *dpy = myDpy;

    XSync(dpy, 0);      /* make certain everything is flushed */
    RETURN ( (XPending(dpy)) ? true : false );
%}
!

eventPendingWithoutSync
    "return true, if any event is pending."

%{  /* UNLIMITEDSTACK */

    RETURN ( (XPending(myDpy)) ? true : false );
%}
!

exposeEventPendingFor:aWindowId
    "return true, if any expose event is pending"

%{  /* UNLIMITEDSTACK */

    Display *dpy = myDpy;
    XEvent ev;
    Window win;

    if (_isSmallInteger(aWindowId)) {
        win = _WindowVal(aWindowId);
        if (XCheckWindowEvent(dpy, win, ExposureMask, &ev)) {
            XPutBackEvent(dpy, &ev);
            RETURN ( true );
        }
    }
    RETURN ( false );
%}
!

eventsPending:anEventMask for:aWindowId
    "return true, if any of the masked events is pending"

%{  /* UNLIMITEDSTACK */

    Display *dpy = myDpy;
    XEvent ev;
    Window win;

    if (_isSmallInteger(anEventMask) && _isSmallInteger(aWindowId)) {
        win = _WindowVal(aWindowId);
        XSync(dpy, 0);      /* make certain everything is flushed */
        if (XCheckWindowEvent(dpy, win, _intVal(anEventMask), &ev)) {
            XPutBackEvent(dpy, &ev);
            RETURN ( true );
        }
    }
    RETURN ( false );
%}
!

eventPending:anEventSymbol for:aWindowId
    "return true, if a specific event is pending"

    ^ self eventsPending:(self eventMaskFor:anEventSymbol) for:aWindowId
! !
